/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.datasource.rdb.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.CalculatedParameterOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.medor.tuple.lib.EmptyTupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.medor.lib.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.sql.*;
import java.util.ArrayList;
import java.io.ObjectInputStream;
import java.io.IOException;


/**
 * This class implements the Wrapper interface for the jdbc resources. This
 * implementation uses Statements to perform the request.
 *
 * @author Mourad Alia
 */
public class JDBCWrapper implements Wrapper {

	/**
	 * this field referencs a Logger which is never null.
	 */
	Logger log = null;
	boolean debug;

	public JDBCWrapper() {
		log = Log.loggerFactory.getLogger(
			Log.MEDOR_PREFIX + ".eval.rdb.JDBCWrapper");
	}


	public TupleCollection fetchData(QueryLeaf lf,
									 ParameterOperand[] parameters,
									 Object conn) throws EvaluationException {
		debug = log.isLoggable(BasicLevel.DEBUG);
		RdbQueryLeaf leaf = (RdbQueryLeaf) lf;
		ArrayList usedParameters = null;
		ResultSet resultSet = null;
		ResultSetMetaData rsmd = null;
		Tuple buffer;
		VariableOperand[] operandBuffer;
		Statement st = null;
		PreparedStatement ps = null;
		if (parameters != null
				&& parameters.length > 0
				&& leaf instanceof FilteredQueryTree) {
			Expression filter = ((FilteredQueryTree) leaf).getQueryFilter();
			if (filter != null)
				usedParameters = getUsedParameters(filter, parameters, null);
		}
		try {
			if (debug) log.log(BasicLevel.DEBUG, "Evaluate the query: " + leaf.getSqlRequest());
			if (usedParameters == null || usedParameters.size() == 0) {
				if (debug) log.log(BasicLevel.DEBUG, "No parameter => use of a Statement");
				st = ((Connection) conn).createStatement(
					ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_READ_ONLY);
				resultSet  = st.executeQuery(leaf.getSqlRequest());
			}
			else {
				if (debug) log.log(BasicLevel.DEBUG, usedParameters.size() + " parameter(s) => use of a PreparedStatement");
				ps = ((Connection) conn).prepareStatement(leaf.getSqlRequest(),
					ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_READ_ONLY);
				for (int i = 0; i < usedParameters.size(); i++) {
					ParameterOperand po = (ParameterOperand) usedParameters.get(i);
					if (debug) log.log(BasicLevel.DEBUG, "Assign parameter " + po.getName());
					switch (po.getType().getTypeCode()) {
						case QTypeTuple.TYPECODE_INT:
						case QTypeTuple.TYPECODE_OBJINT:
							ps.setInt(i + 1, po.getInt());
							break;
						case QTypeTuple.TYPECODE_SHORT:
						case QTypeTuple.TYPECODE_OBJSHORT:
							ps.setShort(i + 1, po.getShort());
							break;
						case QTypeTuple.TYPECODE_BYTE:
						case QTypeTuple.TYPECODE_OBJBYTE:
							ps.setByte(i + 1, po.getByte());
							break;
						case QTypeTuple.TYPECODE_LONG:
						case QTypeTuple.TYPECODE_OBJLONG:
							ps.setLong(i + 1, po.getLong());
							break;
						case QTypeTuple.TYPECODE_DOUBLE:
						case QTypeTuple.TYPECODE_OBJDOUBLE:
							ps.setDouble(i + 1, po.getDouble());
							break;
						case QTypeTuple.TYPECODE_BOOLEAN:
						case QTypeTuple.TYPECODE_OBJBOOLEAN:
							ps.setBoolean(i + 1, po.getBoolean());
							break;
						case QTypeTuple.TYPECODE_FLOAT:
						case QTypeTuple.TYPECODE_OBJFLOAT:
							ps.setFloat(i + 1, po.getFloat());
							break;
						case QTypeTuple.TYPECODE_STRING:
							ps.setString(i + 1, po.getString());
							break;
						case QTypeTuple.TYPECODE_CHAR:
						case QTypeTuple.TYPECODE_OBJCHAR:
							ps.setInt(i + 1, po.getChar());
							break;
						case QTypeTuple.TYPECODE_DATE:
                            java.util.Date d = po.getDate();
                            if (d == null)
                                ps.setNull(i + 1, java.sql.Types.DATE);
                            else
                                ps.setDate(i + 1, new java.sql.Date(d.getTime()));
							break;
						default:
							throw new EvaluationException("The type "
									+ po.getType().getJormName()
									+ " is not supported '"
								  + po.getType().getTypeCode()+"'"
							);
					}
				}
				if (debug) log.log(BasicLevel.DEBUG, "Evaluate the query");
				resultSet = ps.executeQuery();
			}
            if (!resultSet.next()) {
				if (debug) log.log(BasicLevel.DEBUG, "No result");
				if (st != null)
					st.close();
				if (ps != null)
					ps.close();
                return new EmptyTupleCollection(lf.getTupleStructure());
            }
            else {
				if (debug) log.log(BasicLevel.DEBUG, "There is a result");
                rsmd = resultSet.getMetaData();
                operandBuffer = new VariableOperand[rsmd.getColumnCount()];
                for (int cpt = 0; (cpt < lf.getTupleStructure().getSize()); cpt++) {
                    // create and init the buffer
                    operandBuffer[cpt] = new BasicVariableOperand(leaf.getTupleStructure().getField(cpt + 1).getType());
                }
                buffer = new MemoryTuple(operandBuffer);
                return new JDBCTupleCollection(leaf, resultSet, operandBuffer, buffer);
            }
		}
		catch (SQLException sqlexp) {
			throw new EvaluationException(sqlexp);
		}
		catch (EvaluationException ee) {
			throw ee;
		}
		catch (MedorException mexp) {
			throw new EvaluationException(mexp);
		}
	}

	private class JDBCTupleCollection implements TupleCollection {

		private ResultSet resultSet;
		private RdbQueryLeaf leaf;
		private VariableOperand[] operandBuffer;
		private Tuple buffer;

		public JDBCTupleCollection(RdbQueryLeaf lf, ResultSet rs, VariableOperand[] ob,
								   Tuple b) throws MedorException, SQLException {
			if (debug) log.log(BasicLevel.DEBUG, "TupleCollection creation");
			leaf = lf;
			resultSet = rs;
			operandBuffer = ob;
			buffer = b;
			loadTuple();
		}

		public TupleStructure getMetaData() throws MedorException {
			return leaf.getTupleStructure();
		}

		public boolean isLast() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				return resultSet.isLast();
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}

		/**
		 * Moves the cursor down one row from its
		 * current position to the next tuple
		 */
		public boolean next() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.next()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.next()=true");
					loadTuple();
					return true;
				}
				else {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.next()=false");
					return false;
				}
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}


		// Moves the cursor to the previous Tuple of this TupleCollection
		public boolean previous() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.previous()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=true");
					loadTuple();
					return true;
				}
				else {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=false");
					return false;
				}

			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}

		/**
		 * Moves the cursor to the first Tuple of this TupleCollection
		 */
		public void first() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.first()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=true");
					loadTuple();
				}
				else
					throw new MedorException("can't move the cursor to the first row");
			}
			catch (SQLException ex) {
				throw new MedorException(ex);
			}
		}

		/**
		 * Retrieves the current row Number of this TupleCollection
		 */
		public int getRow() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				return resultSet.getRow();
			}
			catch (SQLException ex) {
				throw new MedorException(ex);
			}
		}

		public Object getObject(int numColumn) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "numColumn=" + numColumn);
			try {
                Object o = resultSet.getObject(numColumn);
                return (resultSet.wasNull() ? null : o);
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
			catch (Exception e) {
				throw new MedorException("Invalid field index " + numColumn, e);
			}
		}

		public Tuple getTuple() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			return buffer;
		}

		private void loadTuple() throws MedorException, SQLException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			PType ptype = null;
			for (int cpt = 1; (cpt <= leaf.getTupleStructure().getSize()); cpt++) {
				ptype = leaf.getTupleStructure().getField(cpt).getType();
				switch (ptype.getTypeCode()) {
                case QType.TYPECODE_INT:
                    operandBuffer[cpt - 1].setValue(resultSet.getInt(cpt));
                    break;
                case QType.TYPECODE_SHORT:
                    operandBuffer[cpt - 1].setValue(resultSet.getShort(cpt));
                    break;
                case QType.TYPECODE_BYTE:
                    operandBuffer[cpt - 1].setValue(resultSet.getByte(cpt));
                    break;
                case QType.TYPECODE_LONG:
                    operandBuffer[cpt - 1].setValue(resultSet.getLong(cpt));
                    break;
                case QType.TYPECODE_DOUBLE:
                    operandBuffer[cpt - 1].setValue(resultSet.getDouble(cpt));
                    break;
                case QType.TYPECODE_BOOLEAN:
                    operandBuffer[cpt - 1].setValue(resultSet.getBoolean(cpt));
                    break;
                case QType.TYPECODE_FLOAT:
                    operandBuffer[cpt - 1].setValue(resultSet.getFloat(cpt));
                    break;
                case QType.TYPECODE_STRING:
                    String s = resultSet.getString(cpt);
                    operandBuffer[cpt - 1].setValue((resultSet.wasNull() ? null : s));
                    break;
                case QType.TYPECODE_OBJBOOLEAN:
                    boolean bool = resultSet.getBoolean(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Boolean(bool)));
                    break;
                case QType.TYPECODE_OBJCHAR:
                    String cc = resultSet.getString(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Character(cc.charAt(0))));
                    break;
                case QType.TYPECODE_OBJBYTE:
                    byte by = resultSet.getByte(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Byte(by)));
                    break;
                case QType.TYPECODE_OBJSHORT:
                    short sh = resultSet.getShort(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Short(sh)));
                    break;
                case QType.TYPECODE_OBJINT:
                    int in = resultSet.getInt(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Integer(in)));
                    break;
                case QType.TYPECODE_OBJLONG:
                    long lon = resultSet.getLong(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Long(lon)));
                    break;
                case QType.TYPECODE_OBJFLOAT:
                    float fl = resultSet.getFloat(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Float(fl)));
                    break;
                case QType.TYPECODE_OBJDOUBLE:
                    double dou = resultSet.getDouble(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? null : new Double(dou)));
                    break;
                case QType.TYPECODE_DATE:
                    java.sql.Date d = resultSet.getDate(cpt);
                    operandBuffer[cpt - 1].setValue((resultSet.wasNull() ? null : d));
                    break;
                case QType.TYPECODE_CHAR:
                    String ch = resultSet.getString(cpt);
                    operandBuffer[cpt - 1].setValue(
                        (resultSet.wasNull() ? Character.MIN_VALUE : ch.charAt(0)));
                    break;
                case PType.TYPECODE_SERIALIZED:
                    Blob b = resultSet.getBlob(cpt);
                    if (resultSet.wasNull())
                        operandBuffer[cpt - 1].setValue((Object) null);
                    else
                        try {
                            operandBuffer[cpt - 1].setValue(
                                new ObjectInputStream(
                                    b.getBinaryStream()).readObject());
                        }
                        catch (SQLException e) {
                            throw e;
                        }
                        catch (Exception e) {
                            throw new MedorException(
                                "Impossible to read the serialized field "
                                + leaf.getTupleStructure().getField(cpt).getName(), e);
                        }
                    break;
				}
			}
		}

		public Tuple getTuple(int row) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (debug) log.log(BasicLevel.DEBUG, "resultSet.absolute(" + row + ")");
				if (resultSet.absolute(row)) {
					loadTuple();
				}
				else
					throw new MedorException("Can't move to the row: " + row);
			}
			catch (SQLException sqlexp) {
				throw new MedorException(sqlexp);
			}
			return buffer;
		}

		public boolean isEmpty() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "return " + (buffer == null));
			return (buffer == null);
		}

		public boolean row(int i) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (debug) log.log(BasicLevel.DEBUG, "resultSet.absolute(" + i + ")");
				if (resultSet.absolute(i)) {
					loadTuple();
					return true;
				}
				else
					return false;
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}
	}

	private ArrayList getUsedParameters(Expression e,
										ParameterOperand[] parameters,
										ArrayList al)
			throws EvaluationException {
		if (e instanceof Operator) {
			for(int i=0; i<((Operator) e).getOperandNumber(); i++) {
				al = getUsedParameters(((Operator) e).getExpression(i), parameters, al);
			}
		}
		else if (e instanceof CalculatedParameterOperand) {
			CalculatedParameterOperand cpo = (CalculatedParameterOperand) e;
			cpo.evaluate(parameters);
			if (al == null)
				al = new ArrayList();
			al.add(cpo);
		}
		else if (e instanceof ParameterOperand) {
			int i = 0;
			ParameterOperand param = (ParameterOperand) e;
			while (i < parameters.length
					&& !parameters[i].getName().equals(param.getName()))
				i++;
			if (i < parameters.length) {
                if (parameters[i].getType().getTypeCode() != param.getType().getTypeCode()) {
                    throw new EvaluationException("Bad parameter type: found "
                        + parameters[i].getType().getJormName()
                        + " / expected " + param.getType().getJormName());
                }
				if (al == null)
					al = new ArrayList();
                al.add(parameters[i]);
			}
			else
				throw new EvaluationException(
						"You must specify the value of the parameter "
						+ param.getName());
		}
		return al;

	}
}
