/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.jorm.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.eval.lib.TuplePNameGetter;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.tuple.api.Tuple;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * A CompositePName is an Operator representing the construction of a PName
 * from a NamingContext and several Fields.<br>
 * The NamingContext is provided as a ParameterOperand. The name of the
 * ParameterOperand is defined to be:
 * <ol><li>The fully qualified class name for the PName of the class itself</li>
 * <li>The fully qualified class name, to which the attribute name is
 * concatenated for a reference PName</li>
 * </ol>
 * The Fields are provided as an array of FieldOperand.
 * It relies on a TuplePNameGetter
 * (@see org.objectweb.medor.query.jorm.lib.TuplePNameGetter) to construct the
 * PName.

 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class CompositePName implements Operator {

	protected Map cofn2clfn;
	protected PType type = null;
	protected FieldOperand[] fields = null;
	protected String[] cofns = null;
	protected Expression ncParam = null;
	protected VariableOperand result = null;
	protected boolean verified = false;
	protected TuplePNameGetter myGetter = null;
	protected CompositeName compositeName = null;

	/**
	 * Constructs a SinglePName from a ParameterOperand representing the
	 * naming context, and an array of FieldOperand. The FieldOperands should
	 * be built on JormFields.
	 * @param fs is the array of FieldOperands from which the value is taken.
	 * @param p is the ParameterOperand representing the naming context.
	 * @param nameType is the PType of the PName created by this operator.
	 * @param proj contains the associations between a composite field name and
	 * a jorm class field name.
	 * @throw a MedorException when the number of FieldOperand is not equal to
	 * the size of the proj parameter.
	 */
	public CompositePName(FieldOperand[] fs,
						  ParameterOperand p,
						  PType nameType,
						  Map proj,
						  CompositeName cn,
						  String alias) throws MedorException {
		if (proj.size() != fs.length)
			throw new MedorException("Bad fieldOperand numnber (" + fs.length
				+ ") compared to the map size (" + proj.size() + ")");
		cofn2clfn = proj;
		cofns = new String[proj.size()];
		compositeName = cn;
		this.fields = fs;
		this.ncParam = p;
		this.type = nameType;
		//construct the Map for the TuplePNameGetter
		Map name2Index = new HashMap(fields.length);
		for (Iterator it = proj.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			//System.out.println("\nCompositeName: key=" + (String) me.getKey() + " / value("+me.getValue()+")=" + getFieldIndex(alias + "." + me.getValue(), fields));
			String cofn = (String) me.getKey();
			name2Index.put(cofn,
				getFieldIndex(alias + "." + (String) me.getValue(), fields, cofn));
		}
		//build the TuplePNameGetter
		myGetter = new TuplePNameGetter(name2Index);
	}

	private Integer getFieldIndex(String fn, FieldOperand[] fields, String cofn)
		throws MedorException {
		if (fn == null)
			return null;
		int i = 0;
		while(i<fields.length && !fields[i].getField().getName().equals(fn)) {
			i++;
		}
		if (i < fields.length) {
			cofns[i] = cofn;
			// The tuple indexes start to 1 and not 0
			QueryTreeField qtf = ((QueryTreeField) fields[i].getField());
			return new Integer(qtf.getQueryTree().getTupleStructure().getFieldRank(qtf));
		}
		throw new MedorException("Impossible to find the field " + fn);
	}

	private String getShortName(String fn) {
		int idx = fn.indexOf(".");
		return (idx == -1
			? null
			: fn.substring(idx + 1, fn.length()));
	}

	/**
	 * It retrieves the map which contains the association between composite
	 * field name to class field Name
	 */
	public Map getProjection() {
		return cofn2clfn;
	}

	/**
	 * It retrieves an array of the composite field name. The order of this
	 * array matches to the order of the FieldOperand array returned by the
	 * method getField()
	 */
	public String[] getCompositeFieldName() {
		return cofns;
	}

	public CompositeName getCompositeName() {
		return compositeName;
	}

	// IMPLEMENTATION OF THE Operator INTERFACE //
	//------------------------------------------//

	public int getOperandNumber() {
		return 2;
	}

	public Expression getExpression(int idx)
		throws ArrayIndexOutOfBoundsException {
		if (idx == 0)
			return ncParam;
		else if (idx < (fields.length + 1))
			return fields[idx - 1];
		else
			throw new ArrayIndexOutOfBoundsException(idx);
	}

	public void setExpression(int idx, Expression exp)
		throws ArrayIndexOutOfBoundsException {
		if (idx == 0)
			ncParam = exp;
		else if (idx < (fields.length + 1))
			fields[idx - 1] = (FieldOperand) exp;
		else
			throw new ArrayIndexOutOfBoundsException(idx);
	}

	// IMPLEMENTATION OF THE Expression INTERFACE //
	//--------------------------------------------//

	public PType getType() {
		return type;
	}

	/**
	 * CHecks wether this expression is compiled or not
	 * @return true if the expression is compiled and false if it is not
	 */
	public boolean isCompiled() {
		return verified;
	}

	//TODO
	public Object clone() throws CloneNotSupportedException {
		return null;
	}

	/**
	 * Gets the operand in wich the result will be put in
	 * @return an <code>Operand</code> Object
	 * @throws <code>IllegalStateException</code> if this expression is not
	 * compiled.
	 */
	public Operand getResult() throws IllegalStateException {
		if (verified) {
			return result;
		} else
			throw new IllegalStateException("Can't get result of an uncompiled expression");
	}

	/**
	 *
	 */
	public void evaluate(ParameterOperand[] pos, Tuple t)
		throws MedorException {
		//assign the value of the NamingContext
		ncParam.evaluate(pos, t);
		//gets the naming context
		if (result == null)
			result = new BasicVariableOperand(type);
		PNamingContext nc =
			(PNamingContext) ((VariableOperand) ncParam).getObject();
		try {
			myGetter.setTuple(t);
			result.setValue(nc.decodeAbstract(myGetter, null));
		} catch (PException pe) {
			throw new MedorException("Problem for builing composite PName from tuple", pe);
		}
	}

	/**
	 * Checks the semantic integrity of an expression.
	 * It checks that all types are compatible and prepare the expression to be
	 * evaluable. It also creates buffers where stores the result. Notes that
	 * when evaluating there is no creation of new objects. This method changes
	 * the state of this expression object, it will be evaluable and not
	 * modifiable.
	 * @throws <code>TypingException</code> when incompatible types error
	 * occures.
	 * @throws <code>MalformedExpressionException</code> if syntax error
	 */
	public void compileExpression()
		throws TypingException, MalformedExpressionException {
		if ((fields != null) && (ncParam != null)) {
			result = new BasicVariableOperand(type);
			verified = true;
		} else
		// Operands not yet assigned
			throw new MalformedExpressionException("null children value");
	}


	public FieldOperand[] getFields() {
		return fields;
	}

	//Implementation of Operator interface
	/**
	 * There is no Java operator associated to a PName
	 */
	public String getOperatorString() {
		return "compositePName";
	}
}
