/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.optim.lib.BasicRule;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Map;

/**
 * It permits to assign the mappers to the jorm leaves. This rule must be execute
 * before the rewriting of the jorm leaves. A map must be assigned to this rule.
 * This map has the following caracteristics:
 * key = jorm class name
 * value = the mapper instance which maps the jorm class.
 *
 * Currently this rule does not take in account the possibility to map the same
 * class several times into several mapper.
 *
 * @author S.Chassande-Barrioz
 */
public class JormAssignMapperRule extends BasicRule {

    protected Map jormName2Mapper = null;

	public JormAssignMapperRule() {
		super("JormAssignMapperRule");
	}


    public Map getJormName2Mapper() {
        return jormName2Mapper;
    }

    /**
     * It assignes the map used in this rule. (Read the class description for
     * more details).
     * @param jormName2Mapper is the map
     */
    public void setJormName2Mapper(Map jormName2Mapper) {
        this.jormName2Mapper = jormName2Mapper;
    }

    public QueryTree rewrite(QueryTree qt) throws MedorException {
        if (jormName2Mapper == null || jormName2Mapper.size()==0)
            throw new MedorException("Unable to assign the mapper on the jorm leaves without non empty map");
        if (debug)
            log.log(BasicLevel.DEBUG, "Assignation of mappers");
        assignMapper(qt);
        return qt;
    }

    protected void assignMapper(QueryTree qt) throws MedorException {
        if (qt instanceof ClassExtent) {
            ClassExtent extent = (ClassExtent) qt;
            String name = extent.getMetaObject().getFQName();
            PMapper mapper = (PMapper) jormName2Mapper.get(name);
            if (mapper != null) {
                extent.setPMapper(mapper);
                if (debug)
                    log.log(BasicLevel.DEBUG, "Assign the mapper "
                        + mapper.getMapperName() + "to the ClassExtent" + extent.getName());
            } else
                throw new MedorException("No mapper specified for the class " + name);
        }
        else if (qt instanceof GenClassExtent) {
            GenClassExtent gce = (GenClassExtent) qt;
            GenClassRef gcr = gce.getMetaObject();
            String name = ((Class) gcr.getParent()).getFQName() + "." + gcr.getName();
            PMapper mapper = (PMapper) jormName2Mapper.get(name);
            if (mapper != null) {
                gce.setPMapper(mapper);
                if (debug)
                    log.log(BasicLevel.DEBUG, "Assign the mapper "
                        + mapper.getMapperName() + "to the GenClassExtent" + gce.getName());
            } else
                throw new MedorException("No mapper specified for the GenClass " + name);
        }
        else if (qt instanceof QueryNode) {
            QueryTree[] qts = ((QueryNode) qt).getChildren();
            for (int i = 0; i < qts.length; i++) {
                assignMapper(qts[i]);
            }
        }
    }
}
