/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.optim.api.LeafRewriter;
import org.objectweb.medor.optim.lib.BasicRule;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.util.Map;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class JormLeafRewriter extends BasicRule implements LeafRewriter {

	public JormLeafRewriter() {
		super();
	}
	public JormLeafRewriter(String suffix) {
		super(suffix);
	}

	public abstract boolean canRewrite(QueryLeaf ql);

	public QueryTree rewrite(QueryLeaf ql) throws MedorException {
		if (!canRewrite(ql))
			throw new MedorException(
				"Impossible to rewrite this QueryLeaf: " + ql);
		ClassExtent extent = (ClassExtent) ql;
		String mapperName = extent.getPMapper().getMapperName();
		String mappingName = extent.getPMapper().getMappingName();
		Class clazz = extent.getMetaObject();
        //System.out.println("Class:" + clazz.getName());
		QueryLeaf specificQL = createQueryLeaf(extent, mapperName, mappingName);
		QueryNode qt = new Project("", null);

		ArrayList fieldNames = new ArrayList();
		Field[] oldFields = ql.getTupleStructure().getFields();
		boolean classpnametreated = false;
		for (int i = 0; i < oldFields.length; i++) {
			String fieldName = oldFields[i].getName();
            //System.out.println("FielName:" + fieldName);
			TypedElement te = (TypedElement)
                clazz.getTypedElement(getShortFieldName(fieldName));
			if (te == null) {
                //System.out.println("PName of the class");
                // PName of the class
				if (classpnametreated) {
					throw new MedorException("Unknown field: " + fieldName);
                }
				else {
					classpnametreated = true;
					addReference(clazz.getNameDef(mappingName),
						oldFields[i].getName(),
						oldFields[i].getType(),
						qt,
						specificQL,
						fieldNames,
						clazz,
						mapperName,
						mappingName);
				}
			}
			else if (te instanceof PrimitiveElement) {
                //System.out.println("PrimitiveElement: " + te.getName());
				String fn = getFieldName(specificQL.getName(), te.getName());
                QueryTreeField qtf = null;
                if (!fieldNames.contains(fn)) {
                    //Register the fieldName
                    fieldNames.add(te.getName());
                    // add the field into the QueryLeaf
                    qtf = addPrimitiveElement(specificQL,
                        (PrimitiveElement) te,
                        mapperName,
                        mappingName);
                }
                else {
                    qtf = (QueryTreeField)
                        specificQL.getTupleStructure().getField(fn);
                }
                //add a propagated field into the intermediate QueryTree
				qt.addPropagatedField(
                    qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});

			}
			else if (te instanceof Reference) {
                //System.out.println("Reference: " + te.getName());
				addReference(((Reference) te).getNameDef(mappingName),
					oldFields[i].getName(),
					oldFields[i].getType(),
					qt,
					specificQL,
					fieldNames,
					clazz,
					mapperName,
					mappingName);
			}
            else
                throw new MedorException("Unknown field: " + te.getName());
		}
		return qt;
	}

	protected abstract QueryLeaf createQueryLeaf(ClassExtent extent,
												 String mapperName,
												 String mappingName)
		throws MedorException;

	public abstract QueryTreeField addPrimitiveElement(QueryLeaf neo,
													   PrimitiveElement pe,
													   String mapperName,
													   String mappingName)
		throws MedorException;

	protected PrimitiveElement getPrimitiveElement(Class clazz,
												   String fieldName)
		throws MedorException {
        String fn = getShortFieldName(fieldName);
		PrimitiveElement pe = (PrimitiveElement)
			clazz.getTypedElement(fn);
		if (pe == null) {
			pe = clazz.getHiddenField(fn);
            if (pe == null)
                throw new MedorException("No field " + fn
				+ "(" + fieldName + ") found the the class " + clazz.getName());
        }
		return pe;
	}

	protected void addReference(NameDef nd,
								String name,
								PType type,
								QueryNode qt,
								QueryLeaf specificQL,
								ArrayList fieldNames,
								Class clazz,
								String mapperName,
								String mappingName) throws MedorException {
		QueryTreeField qtf = null;
		if (nd.isFieldName()) {
			String ndfn = getFieldName(specificQL.getName(), nd.getFieldName());
			//Check if the field has not already added on the specific query leaf
			PrimitiveElement pe = getPrimitiveElement(clazz, ndfn);
			if (!fieldNames.contains(ndfn)) {
				fieldNames.add(ndfn);
				// add the field into the QueryLeaf
				qtf = addPrimitiveElement(specificQL, pe, mapperName, mappingName);
            } else {
				// Fetch the existing field on the specific QueryLeaf
				qtf = (QueryTreeField) specificQL.getTupleStructure()
					.getField(getFieldName(specificQL.getName(), pe.getName()));
			}

			//add a CalculatedField into the intermediate QueryTree.
			// This calculatedField has an expression which is a
			// SinglePName.
			Expression e = new SinglePName(
				new BasicFieldOperand(qtf),
				new BasicParameterOperand(PTypeSpaceMedor.NAMING_CONTEXT,
					clazz.getFQName() + "." + getShortFieldName(name)),
				type);
			qt.addCalculatedField(name, type, e);
		}
		else if (nd.isNameRef()) {
            Map proj = nd.getNameRef().getProjection();
			Collection c = proj.values();
			FieldOperand[] fos = new BasicFieldOperand[c.size()];
			int fosCounter = 0;
			for (Iterator it = c.iterator(); it.hasNext();) {
				String ndfn =
                    getFieldName(specificQL.getName(),(String) it.next());
				if (!fieldNames.contains(ndfn)) {
					fieldNames.add(ndfn);
					qtf = addPrimitiveElement(specificQL,
						getPrimitiveElement(clazz, ndfn),
						mapperName,
						mappingName);
				}
				else {
					qtf = (QueryTreeField)
						specificQL.getTupleStructure().getField(ndfn);
				}
				fos[fosCounter++] = new BasicFieldOperand(qtf);
			}
			//add a CalculatedField into the intermediate QueryTree.
			// This calculatedField has an expression which is a
			// CompositePName.
			Expression e = new CompositePName(
				fos,
				new BasicParameterOperand(PTypeSpaceMedor.NAMING_CONTEXT,
										  clazz.getFQName() + "." + getShortFieldName(name)),
				type,
                proj,
                nd.getNameRef().getCompositeName(),
				specificQL.getName());
			qt.addCalculatedField(name, type, e);
		}
	}
    protected String getFieldName(String prefix, String nodeName) {
        if (prefix==null || prefix.length()==0)
            return nodeName;
        else
            return prefix + "." + nodeName;
    }

    protected String getShortFieldName(String fn) {
        return fn.substring(fn.lastIndexOf(".")+1, fn.length());
    }
}
