/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.rdb.RDBClassMapping;
import org.objectweb.jorm.metainfo.api.rdb.RDBColumnSpec;
import org.objectweb.jorm.metainfo.api.rdb.RDBValueMapping;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.util.api.Loggable;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.datasource.lib.ConnectionFactoryDataStore;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;

/**
 * @author S.Chassande-Barrioz
 */
public class RdbJormLeafRewriter extends JormLeafRewriter {

	public final static String MAPPER_NAME = "rdb";

	public RdbJormLeafRewriter() {
		super("RdbJormLeafRewriter");
	}

	public boolean canRewrite(QueryLeaf ql) {
		if (!(ql instanceof ClassExtent))
			return false;
		ClassExtent extent = (ClassExtent) ql;
		PMapper mapper = extent.getPMapper();

		if (mapper == null) {
			if (debug)
				log.log(BasicLevel.ERROR,
						"No mapper associated to the QueryLeaf " + ql.getName());
			return false;
		}
		String mapperName = mapper.getMapperName();
		String mappingName = mapper.getMappingName();
		if (!MAPPER_NAME.equals(mapperName)
			&& !mapperName.startsWith(MAPPER_NAME + "."))
			return false;

		ClassMapping cm =
				extent.getMetaObject().getClassMapping(mapperName, mappingName);
		return cm != null && cm instanceof RDBClassMapping;
	}


	public QueryLeaf createQueryLeaf(ClassExtent extent,
									 String mapperName,
									 String mappingName) throws MedorException {
		RDBClassMapping cm = (RDBClassMapping)
				extent.getMetaObject().getClassMapping(mapperName, mappingName);

		BasicQualifiedTable qt = new BasicQualifiedTable(cm.getTableName(), null);
		DataStore ds = extent.getDataStore();
		if (ds == null) {
			Object cf = extent.getPMapper().getConnectionFactory();
			ds = new ConnectionFactoryDataStore(
					DataStore.JDBC_STORE,
					cf.toString(),
					new short[0],
					cf);
		}
		RdbExpQueryLeaf neo = new BasicRdbExpQueryLeaf(
				ds,
				new BasicQualifiedTable[]{qt},
				extent.getName());
		return neo;
	}

	public QueryTreeField addPrimitiveElement(QueryLeaf neo,
											  PrimitiveElement pe,
											  String mapperName,
											  String mappingName) throws MedorException {
		return ((RdbExpQueryLeaf) neo).addRdbField(
				getFieldName(neo.getName(), pe.getName()),
				pe.getType(),
				getColumnName(pe, mapperName, mappingName),
				((RdbExpQueryLeaf) neo).getQualifiedTables()[0]);
	}

	private String getColumnName(PrimitiveElement pe,
								 String mapperName,
								 String mappingName) throws MedorException {
		RDBValueMapping vm = (RDBValueMapping)
				pe.getValueMapping(mapperName, mappingName);
		if (vm == null)
			throw new MedorException(
					"No ValueMapping (" + mapperName + ", " + mappingName
					+ ") availlable for the field: " + pe.getName());
		Iterator it = vm.iterateRDBColumnSpec();
		if (!it.hasNext())
			throw new MedorException(
					"No RDBColumnSpec availlable for the field: " + pe.getName());
		return ((RDBColumnSpec) it.next()).getName();
	}
}
