/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.rdb;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class groups together as a single RbdExpQueryLeaf the RdbQueryLeaves
 * working on the same data store. This rule does not reorganize a QueryTree
 * to put close together leaves working on the same data store; this should be
 * done by another rule.
 */
public class GroupSameDBRule implements RewriteRule {
	/**
	 * The RdbQuery class represents the result of rewriting a QueryTree.
	 * It has an array of RdbExpFields (its fields), an array of QualifiedTables
	 * (its tables), an Expression working on those Fields, an associated
	 * DataStore and a boolean indicating whether it is a query leaf, or whether
	 * it is the result of merging several QueryLeaves.
	 */
	public class RdbQuery {
		public ArrayList fields;
		public ArrayList tables;
		public Expression exp;
		public DataStore ds;
		public boolean isLeaf;
		//tableIds is the current list of table names (without alias) and
		//table alias names.
		public ArrayList tableIds;
	}

	public QueryTree rewrite(QueryTree qt) throws MedorException {
		RdbQuery rq = recurseRewrite(qt);
		debug("end of recurse");
		if (rq != null) {
			debug("building RdbLeaf");
			return buildRdbLeaf(rq, qt.getName());
		}
		else
			return qt;
	}

	/**
	 * recurseRewrite calls the transformation on a QueryTree, recursively
	 * on its children.
	 */
	private RdbQuery recurseRewrite(QueryTree qt) throws MedorException {
		//debug(qt.getName());
		if (qt instanceof RdbExpQueryLeaf) {
			debug("RdbExpQueryLeaf: " + qt.getName());
			RdbQuery rq = new RdbQuery();
			Field[] lf = qt.getTupleStructure().getFields();
			rq.fields = new ArrayList(lf.length);
			for (int i = 0; i < lf.length; i++) {
				rq.fields.add(lf[i]);
			}
			QualifiedTable[] lqt = ((RdbExpQueryLeaf)qt).getQualifiedTables();
			rq.tables = new ArrayList(lqt.length);
			rq.tableIds = new ArrayList(lqt.length);
			for (int i = 0; i < lqt.length; i++) {
				rq.tables.add(lqt[i]);
				if (lqt[i].getAliasName() != null)
					rq.tableIds.add(lqt[i].getAliasName());
				else
					rq.tableIds.add(lqt[i].getTableName());
			}
			rq.exp = ((RdbExpQueryLeaf)qt).getQueryFilter();
            rq.ds = ((RdbExpQueryLeaf)qt).getDataStore();
			rq.isLeaf = true;

			debug("RdbExpQueryLeaf: END " + qt.getName());
			return rq;
		}
		else if (qt instanceof QueryLeaf) {
			debug("QueryLeaf: " + qt);
			return null;
		}
		else { //qt is a QueryNode
			debug("QueryNode: " + qt);
			QueryTree[] children = ((QueryNode)qt).getChildren();
			boolean first = true;
			boolean sameStore = true;
			RdbQuery workRq = null;
			//recurse on the Children
			for (int i = 0; i < children.length; i++) {
				RdbQuery currentRq = recurseRewrite(children[i]);
				if (first) {
					debug("first");
					workRq = currentRq;
					first = false;
				}
				else if (sameStore &&
						 workRq.ds != null &&
						 workRq.ds.isSameAs(currentRq.ds)) { //same datastore
					debug("same store => merging");
					mergeRdbQueries(workRq, currentRq);
				}
				else {
					throw new MedorException(
						"Impossible to group two nodes with different data stores");
/*
// This code is WRONG
					debug("different store");
					debug("workRq.ds=" + workRq.ds);
					if (workRq.ds != null) {
						debug("workRq.ds.getDataStoreType()=" + workRq.ds.getDataStoreType());
						debug("workRq.ds.getName()=" + workRq.ds.getName());
					}
					debug("currentRq.ds=" + currentRq.ds);
					if (currentRq.ds != null) {
						debug("currentRq.ds.getDataStoreType()=" + currentRq.ds.getDataStoreType());
						debug("currentRq.ds.getName()=" + currentRq.ds.getName());
					}
					RdbExpQueryLeaf qlWork, qlCurrent;
					//different data stores: process each separately
					//creates a node for workRq if possible
					if (workRq != null)
						qlWork = buildRdbLeaf(workRq, null);
					//creates a node for currentRq if possible
					if (currentRq != null)
						qlCurrent = buildRdbLeaf(workRq, null);
					//sets the workRq to false
					workRq = null;
					sameStore = false;
*/
				}
			}
			if (noCalculation(qt)) {
				debug("noCalculation: => continue");
				//add exp of current node to the workRq
				workRq.exp =
					new And(workRq.exp, ((QueryNode)qt).getQueryFilter());
			}
			else {
				debug("has calculated field: => stop, create RdbExpQL and link");
				RdbExpQueryLeaf ql = buildRdbLeaf(workRq, qt.getName());
				// Upgrade the link between qt and the workRq

				//update fields
				Field[] fs = qt.getTupleStructure().getFields();
				TupleStructure ts = ql.getTupleStructure();
				for (int i=0; i< fs.length; i++) {
					if (fs[i] instanceof PropagatedField) {
						if (ts.contains(fs[i].getName())) {
							debug("Replace in the PropagatedField " + fs[i].getName());
							((QueryNode) qt).updatePropagatedField(
								fs[i].getName(),
								new QueryTreeField[] {
									(QueryTreeField) ts.getField(fs[i].getName())});
						}
						else
							debug("Do not replace in the PropagatedField " + fs[i].getName());
					} else if (fs[i] instanceof CalculatedField) {
						debug("Try to replace in the CalculatedField " + fs[i].getName());
						Expression e = ((CalculatedField) fs[i]).getExpression();
						if (replaceInFilter(e, ts))
							((QueryNode) qt).updateCalculatedField(fs[i].getName(), e);
					} else if (fs[i] instanceof NestedField) {
						debug("Replace in the NestedField " + fs[i].getName());
						throw new MedorException("Unmanaged Nested Field");
					}
				}

				//update filter
				if (qt instanceof FilteredQueryTree) {
					debug("Try to replace in the filter ");
					Expression e = ((FilteredQueryTree) qt).getQueryFilter();
					if (replaceInFilter(e, ts))
						((FilteredQueryTree) qt).setQueryFilter(e);
				}

				workRq = null;
			}
			debug("QueryNode: END");
			return workRq;
		}
	}

	/**
	 * It replaces the fields referenced in the expression by the fields of the
	 * tuple structure. This the field name that is used to match the fields
	 * @param e is the expression where the Field referenced by FieldOperand
	 * must be replaced.
	 * @param ts is the tuple structure that contains the fields that replace
	 * the expression fields.
	 * @return true if the expression was modified, otherwise false.
	 * @throw a MedorException when the data structure is wrong.
	 */
	protected boolean replaceInFilter(Expression e, TupleStructure ts)
		throws MedorException {
		ArrayList al = new ArrayList();
		al.add(e);
		boolean isModified = false;
		while (!al.isEmpty()) {
			Expression ex = (Expression) al.remove(0);
			if (ex instanceof Operator) {
				for (int i = 0; i < ((Operator) ex).getOperandNumber(); i++) {
					al.add(((Operator) ex).getExpression(i));
				}
			} else if (ex instanceof FieldOperand) {
				FieldOperand fo =  (FieldOperand) ex;
				if (ts.contains(fo.getField().getName())) {
					debug("Replace in the FieldOperand: " + fo.getField().getName());
					fo.setField(ts.getField(fo.getField().getName()));
					isModified = true;
				}
				else
					debug("Do not replace in the FieldOperand: " + fo.getField().getName());
			}
		}
		return isModified;
	}


	/**
	 * A QueryTree is elligible for being rewritten as an RdbQueryLeaf if all
	 * its children have been rewritten, and if it does not have any calculated
	 * fields.
	 */
	private boolean noCalculation(QueryTree qt) {
		debug("Testing for calculated fields: ");
		Field[] qtf = qt.getTupleStructure().getFields();
		for (int i = 0; i < qtf.length; i++) {
			if (qtf[i] instanceof CalculatedField) {
				debug("found");
				return false;
				}
		}
		debug("not found");
		return true;
	}

	/**
	 * mergeDataStores merges two RdbQueries. It has to rename the
	 * QualifiedTables if they have the same name.
	 */
	private void mergeRdbQueries(RdbQuery into, RdbQuery other) {
		for (int i = 0; i < other.fields.size(); i++) {
			into.fields.add(other.fields.get(i));
		}
		QualifiedTable qt;
		for (int i = 0; i < other.tables.size(); i++) {
			qt = (QualifiedTable)other.tables.get(i);
			if (into.tables.contains(qt))  {
				//the actual same object is used in another query tree. It needs to
				//be replaced, and all references to qt in the fields
				QualifiedTable qt1 =
					new BasicQualifiedTable(qt.getTableName(),
											qt.getAliasName());
				for (int j = 0; j < other.fields.size(); j++) {
					if (((RdbExpField)other.fields.get(j)).getTable() == qt)
						((RdbExpField)other.fields.get(j)).setTable(qt1);
				}
				qt = qt1;
			}
			//checks aliases: if it has an alias, that alias must not already
			//exist
			if (qt.getAliasName() != null) {
				if (into.tableIds.contains(qt.getAliasName()))
					qt.setAliasName(generateAlias(into.tableIds,
													qt.getAliasName()));
				}
			else if (into.tableIds.contains(qt.getTableName()))
					qt.setAliasName(generateAlias(into.tableIds,
													qt.getTableName()));
			into.tables.add(qt);
		}
		if (into.exp == null) //into.exp is null. Take other.exp
			into.exp = other.exp;
		else if (other.exp != null) //into.ext and other.exp are not null
			into.exp = new And(into.exp, other.exp);
		into.isLeaf = false;
	}

	private String generateAlias(ArrayList tableIds, String root) {
		root += "_";
		int i = 1;
		while (tableIds.contains(root+i)) {
			i++;
			}
		root += i;
		tableIds.add(root);
		return root;
	}

	/**
	 * buildRdbLeaf constructs an RdbExpQueryLeaf from the result of grouping
	 * the RdbExpQueryLeaves of a QueryTree, expressed as an RdbQuery.
	 * @param rq is the RdbQuery obtained when grouping the RdbExpQueryLeaves
	 * of the QueryTree
	 * @param qt is the QueryTree for which the RdbExpQueryLeaves have been
	 * grouped.
	 */
	private RdbExpQueryLeaf buildRdbLeaf(RdbQuery rq, String name)
		throws MedorException {
		RdbExpQueryLeaf ql =
			new BasicRdbExpQueryLeaf(
				rq.ds,
				(QualifiedTable[])rq.tables.toArray(new QualifiedTable[0]),
				name);
		ql.setQueryFilter(rq.exp);
		for (int i = 0; i < rq.fields.size(); i++) {
			ql.addRdbField((RdbExpField)rq.fields.get(i));
		}
		return ql;
	}

	private static final void debug(String msg) {
		//System.out.println(msg);
	}
}
