/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.BasicNestedField;
import org.objectweb.medor.lib.BasicTupleStructure;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * This class represents the extent of a JORM class.
 * It contains and implements its associated TupleStructure.
 */
public class ClassExtent extends BasicJormExtent {

	private Class myClass;


	/**
	 * Constructs an extent for the JORM class, in the form of a QueryLeaf.
	 * Two options are possible, depending on the value of the classPNameOnly
	 * parameter:
	 * <ol><li>either all attributes of the JORM class are present in the
	 * QueryLeaf</li>
	 * <li>or only the PName attribute is present in the QueryLeaf</li>
	 * </ol>
	 * @param jormClass is the JORM full class name.
     * @param mapper is the mapper inside which the jorm class will be mapped
     * @param _name is the name of the node (null value => empty string)
	 * @param classPNameOnly indicates whether the extent should contain only
	 * the PName for the class (if true) or all attributes for that class
	 * (if false).
	 *
	 */
	public ClassExtent(Class jormClass, String _name,
                      String pnameFieldName,
					  boolean classPNameOnly) throws MedorException {
        super(_name, pnameFieldName);
		myClass = jormClass;

		//First, create a Field for that Class
		PNameField nameField = new PNameField(getFieldName(name, pnameFieldName), myClass, this);
		this.fields.add(nameField);
		name2field.put(nameField.getName(), nameField);

		if (!classPNameOnly) {

			//Fields are PrimitiveElement, ClassRef or GenClassRef
            Field basicF;
            for (Iterator it = jormClass.iterateAllField(); it.hasNext();) {
				addField((TypedElement) it.next());
			}
		}
	}

	/**
	 * Constructs an extent for the JORM class, in the form of a QueryLeaf.
	 * The field names for those fields which are part of the extent are
	 * passed as a parameter. The fact that the PName is part of the extent
	 * is optional.
	 * @param jormClass is the JORM full class name.
     * @param mapper is the mapper inside which the jorm class will be mapped
     * @param _name is the name of the node (null value => empty string)
	 * @param fieldNames is an array of JORM field names which should be present
	 * in the extent
	 * @param addPName is a boolean which is true if the PName should be part
	 * of the extent, false otherwise.
	 * @throws a MedorException if one of the requested fieldNames does not
	 * correspond to a field of the JORM class.
	 */
	public ClassExtent(Class jormClass, String _name,
                      String[] fieldNames,
					  boolean addPName,
                      String pnameFieldName)
		throws MedorException {
        super(_name, pnameFieldName);
		myClass = jormClass;

		//First, create a Field for that Class if requested
		if (addPName) {
			PNameField nameField = new PNameField(getFieldName(name, pnameFieldName), myClass, this);
			this.fields.add(nameField);
			name2field.put(nameField.getName(), nameField);
		}
		//for each requested fieldName, adds the corresponding
		for (int i = 0; i < fieldNames.length; i++) {
            TypedElement te = jormClass.getTypedElement(fieldNames[i]);
            if (te == null)
				throw new MedorException("No field named " + fieldNames[i] + " in JORM class " + jormClass.getName());
            addField(te);
		}
	}

    /**
     * It adds a new Field into the tuple structure.
     * @param te is the jorm Meta object TypedElement which is the source of
     * the field
     * @param jormClass is the jorm meta object Class associated to this Extent
     * @param pnameFieldName is a string defined by the user to name the class
     * pname or the gen class pname.
     */
    public QueryTreeField addField(TypedElement te) throws MedorException {
        QueryTreeField basicF = null;
		String fn = getFieldName(name, te.getName());
		basicF = (QueryTreeField) name2field.get(fn);
		if (basicF != null)
			throw new MedorException("the field " + te.getName()
				+ " is already defined (" + fn + ")");
        if (te instanceof Reference)
            basicF = new PNameField(fn, myClass, this);
        else
            basicF = new BasicJormField(fn, this, te);
        fields.add(basicF);
        name2field.put(basicF.getName(), basicF);
		return basicF;
    }

	public Class getMetaObject() {
		return myClass;
	}
}
