/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.api.MedorException;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * This class represents Fields corresponding to the PName of a class or a
 * Genclass(the NameDef of the class itself) or the the PName associated to a
 * reference field or to a genclass element.
 * The name of such a Field is a constant.
 *
 * @author Alexandre Lefebvre, Sebastien Chassande-Barrioz
 */
public class PNameField implements QueryTreeField {

    /**
     * In case where this PNameField represents the PName of a Jorm class, this
     * field is not null and references the Jorm Meta object Class.
     */
    protected Class clazz;

    /**
     * In case where this PNameField represents the a reference field, this
     * field is not null and references the Jorm Meta object Reference.
     */
    protected Reference reference;

    /**
     * In case where this PNameField represents GenClass identifier or the
     * genclass element this field is not null and references the Jorm meta
     * object GenClassRef.
     */
    //protected GenClassRef gcr;

    /**
     * If the current PNameField represents a PName of a Jorm class or genclass,
     * this field is equals to 'true'. Otherwise the current PNameField
     * represents a reference field or the genclass element then this field is
     * equals to 'false'.
     */
    protected boolean isClassPName;

    /**
     * This fiedl returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    protected boolean isInGenClass;


    /**
     * Type of the field
     */
    protected PType type;
    protected short valueState;

    /**
     * Name of the field.
     */
    protected String fieldName = null;
    protected QueryTree queryTree = null;

    /**
     * This constructor must be used to built a PNameField associated to a
     * GenClass identifier or the genclass element.
     * @param fieldName is the name which must be given to the field.
     * @param gcr is the jorm meta object which describes the genclass.
     * @param isGenClassPName indicates if this PNameField will represent a
     * genclass identifier(value='true'). Otherwise the this PNameField will
     * represent a genclass element(value='false').
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fieldName,
                      GenClassRef gcr,
                      boolean isGenClassPName,
                      QueryTree qt) {
        this.fieldName = fieldName;
        queryTree = qt;
        this.reference = gcr;
        isInGenClass = true;
        isClassPName = isGenClassPName;
        type = gcr.getType();
    }

    /**
     * This constructor must be used to built a PNameField associated to a Class
     * identifier or a reference field.
     * @param fieldName is the name which must be given to the field.
     * @param cl is the jorm meta object Class associated to this PNameField. In
     * case of the class contains a field which the name is equals to @fn then
     * this PNameField represents a reference field.
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fn, Class cl, QueryTree qt) throws MedorException {
        isInGenClass = false;
        queryTree = qt;
        fieldName = fn;
        TypedElement te = getTypedElement(cl, fieldName);
        //System.out.println("PNameField("+fn + ") te=" + te);
        if (te!=null && te instanceof Reference) {
            type = te.getType();
            isClassPName = false;
            reference = (Reference) te;
        }
        else {
            clazz = cl;
            isClassPName = true;
            MetaObject tmp = clazz.getParent();
            MetaObject par = null;
            while (tmp != null) {
                par = tmp;
                tmp = tmp.getParent();
            }

            try {
                if (clazz.getInheritedClassNumber() == 0) {
                    // no super class, just use the name of the class.
                    type = ((Manager) par)
                        .getPTypeSpace().createPType(clazz.getFQName());
                }
                else {
                    // minimum 1 super class
                    Hashtable table = new Hashtable();
                    clazz.addAllInheritedClass(table);
                    String allNames [][] = new String[table.size()][];
                    int count = 0;

                    Iterator iter = table.values().iterator();
                    while (iter.hasNext()) {
                        String[] names = (String[]) iter.next();
                        allNames[count][0] = names[0];
                        allNames[count] = names;
                    }
                    type = ((Manager) par)
                        .getPTypeSpace().createPType(clazz.getFQName(), allNames);
                }
            }
            catch (Exception e) {
                throw new MedorException(
                    "Impossible to built the type of the class "
                    + clazz.getName(), e);
            }
        }
    }

    /**
     * It returns the type of the field. The type has been calculated a building
     * time.
     */
    public PType getType() {
        return type;
    }

    public short getNullStatus() {
        return valueState;
    }

    public Object clone() {
        return this.clone();
    }

    /**
     * It returns the field name of the PNameField
     */
    public String getName() {
        return fieldName;
    }

    /**
     * It returns 'true' when this PNameField represents the identifier of a jorm
     * class or a jorm genclass.
     */
    public boolean isClassPName() {
        return isClassPName;
    }

    /**
     * It returns the jorm meta object Reference in case of this PNameField
     * represents a reference field in a jorm class.
     */
    public Reference getReference() {
        return reference;
    }

    /**
     * It returns the jorm meta object Class in case of this PNameField
     * represents an identifier of a jorm class.
     */
    public Class getMetaObjectClass() {
        return clazz;
    }

    /**
     * It returns the jorm meta object GenClassRef in case of this PNameField
     * represents a GenClass identifier or a GenClass element.
     */
    public GenClassRef getGenClassRef() {
        return (GenClassRef)reference;
    }

    /**
     * It returns the QueryTree whith which this QueryTreeField is associated.
     */
    public QueryTree getQueryTree() {
        return queryTree;
    }

    /**
     * It returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    public boolean isInGenClass() {
        return isInGenClass;
    }

    private TypedElement getTypedElement(Class clazz, String fn) {
        int idx = fn.length();
        TypedElement te = null;
        te = clazz.getTypedElement(fn);
        while ((idx = fn.lastIndexOf(".", idx)) != -1 && te == null) {
            te = clazz.getTypedElement(fn.substring(idx + 1, fn.length()));
            idx--;
        }
        return te;
    }

}
