/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.NavigatorOperator;
import org.objectweb.medor.query.api.OptimizationMetaData;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;

/**
 * This class constructs a QueryTree from a NavigatorOperator, which represents
 * a path expression.
 * It is a factory for normal JoinProject query nodes, which implement the
 * semantics of the path traversal represented by the NavigatorOperator.
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class NavigatorNodeFactory {
	/**
	 * Creates a QueryTree corresponding to the path traversal represented by
	 * the NavigatorOperator.<br>
	 * The end of the path expression can either be a single-valued field or a
	 * reference to a GenClass. In the latter case, the elements
	 * of the GenClass are projected together with their index attributes.
	 * @param nav is the NavigatorOperator representing the path expression.
	 * This operator starts with a FieldOperand pointing to a Field called here
	 * FStart. All fields in the node hosting FStart are projected.
	 * @param aliasName is the root name of the Fields reached by the path
	 * expression.
	 * @param omd is the optimization meta data related to caching and
	 * parallelism.
	 * @return a new QueryNode which projects all fields in the start QueryTree
	 * and the end QueryTree of the path expression, and which implements the
	 * necessary joins.
	 */
	public static QueryTree getNavigatorNode(NavigatorOperator nav,
											 String aliasName,
											 OptimizationMetaData omd)
		throws MedorException {
		return getField(nav, "object", false, null, aliasName).getQueryTree();
	}

	/**
	 * Creates a QueryTree corresponding to the path traversal represented by
	 * the NavigatorOperator. <br>
	 * The end of the path expression is a reference to a GenClass.
	 * A name is provided for the GenClass element.
	 * @param nav is the NavigatorOperator representing the path expression.
	 * This operator starts with a FieldOperand pointing to a Field called here
	 * FStart. All fields in the node hosting FStart are projected.
	 * @param omd is the optimization meta data related to caching and
	 * parallelism.
	 * @param elemName is the name of the genclass element (an alias) provided
	 * by the user.
	 * @return a new QueryNode which projects all fields in the start QueryTree
	 * and the GenClass element, and which implements the necessary joins.
	 */
	public static QueryTree getIn(NavigatorOperator nav,
								  OptimizationMetaData omd,
								  String elemName)
		throws MedorException {
		return getField(nav, "object", true, elemName, null).getQueryTree();
	}

	/**
	 * Creates a QueryTree corresponding to the path traversal represented by
	 * the NavigatorOperator and a join between the GenClass element
	 * reached by the NavigatorOperator and a given ParameterOperand.<br>
	 * The end of the path expression is a reference to a GenClass.
	 * @param nav is the NavigatorOperator representing the path expression.
	 * This operator starts with a FieldOperand pointing to a Field called here
	 * FStart. All fields in the node hosting FStart are projected.
	 * @param po is the ParameterOperand to be joined with the GenClass element
	 * reached by the path traversal.
	 * @param omd is the optimization meta data related to caching and
	 * parallelism.
	 * @return a new QueryNode which projects all fields in the start QueryTree
	 * and all those in the QueryTree of the FieldOperand, and which implements
	 * the necessary joins.
	 */
	public static QueryTree getMemberOf(NavigatorOperator nav,
								  ParameterOperand po,
								  String elemName,
								  OptimizationMetaData omd)
		throws MedorException {
		QueryTreeField qtf = getField(nav, "object", true, elemName, null);
		//add a SelectProject node
		SelectProject sp = new SelectProject(
			new Equal(po, new BasicFieldOperand(qtf)), "", null);
		Field[] fs = qtf.getQueryTree().getTupleStructure().getFields();
		for (int i = 0; i < fs.length; i++) {
			sp.addPropagatedField(
				fs[i].getName(),
				fs[i].getType(),
				new QueryTreeField[]{(QueryTreeField) fs[i]});
		}
		return sp;
	}
	/**
	 * Creates a QueryTree corresponding to the path traversal represented by
	 * the NavigatorOperator and a join between the GenClass element
	 * reached by the NavigatorOperator and a given FieldOperand.<br>
	 * The end of the path expression is a reference to a GenClass.
	 * @param nav is the NavigatorOperator representing the path expression.
	 * This operator starts with a FieldOperand pointing to a Field called here
	 * FStart. All fields in the node hosting FStart are projected.
	 * @param fop is the FieldOperand to be joined with the GenClass element
	 * reached by the path traversal.
	 * @param omd is the optimization meta data related to caching and
	 * parallelism.
	 * @return a new QueryNode which projects all fields in the start QueryTree
	 * and all those in the QueryTree of the FieldOperand, and which implements
	 * the necessary joins.
	 */
	public static QueryTree getMemberOf(NavigatorOperator nav,
								  FieldOperand fop,
								  String elemName,
								  OptimizationMetaData omd)
		throws MedorException {
		QueryTreeField qtf = getField(nav, "object", true, elemName, null);
		QueryTreeField qtf2 = (QueryTreeField) fop.getField();
		QueryTree qt = qtf.getQueryTree();
		if (qt.getTupleStructure().contains(qtf2.getName())) {
			//Both fields are on the same tree => add a SelectProject node
			qtf2 = (QueryTreeField)
				qt.getTupleStructure().getField(qtf2.getName());
			SelectProject sp  = new SelectProject(
				new Equal(new BasicFieldOperand(qtf2),
					new BasicFieldOperand(qtf)),
				"", null);
			Field[] fs = qt.getTupleStructure().getFields();
			for (int i=0; i< fs.length; i++) {
				sp.addPropagatedField(
					fs[i].getName(),
					fs[i].getType(),
					new QueryTreeField[]{(QueryTreeField) fs[i]});
			}
			return sp;
		} else {
			//The fields are on the different tree => add a JoinProject node
			JoinProject jp  = new JoinProject(
				new Equal(new BasicFieldOperand(qtf2),
					new BasicFieldOperand(qtf)),
				"", null);
			Field[] fs = qt.getTupleStructure().getFields();
			for (int i=0; i< fs.length; i++) {
				jp.addPropagatedField(
					fs[i].getName(),
					fs[i].getType(),
					new QueryTreeField[]{(QueryTreeField) fs[i]});
			}
			return jp;
		}
	}

	/**
	 * This recursive method returns the QueryTreeField which matches to a
	 * NavigatorOperator.
	 * @param nav is the current NavigatorOperator
	 * @param pnameFieldName is the name of the PNameField which could be
	 * created for the class or gen class extent.
	 * @param genClassElemOnly is true if only the GenClass element is
	 * projected, false if all GenClass indexes attributes are projected too.
	 * @param genClassElemName is the name of the GenClass element in the
	 * result QueryTree. If void, the original name is taken.
	 * @param aliasName is the prefix name which all fields of the end of the
	 * path will have.
	 * @return The field can be a PNameField or a propagated field which
	 * references a PNameField.
	 */
	private static QueryTreeField getField(NavigatorOperator nav,
										   String pnameFieldName,
										   boolean genClassElemOnly,
										   String genClassElemName,
										   String aliasName)
		throws MedorException {
		// 'qtf' is the QueryTreeField to use for a join
		QueryTreeField qtf;
		// pnf is the PNameField which must be used to access to the meta information
		PNameField pnf;

		String fieldName = ((Operand) nav.getRightExpression()).getString();

		Expression e = nav.getLeftExpression();
		//System.out.println("nav(" + e + "," + fieldName+")");
		boolean propagQtf;
		if (e instanceof FieldOperand) {
			qtf = (QueryTreeField) ((FieldOperand) e).getField();
			pnf = getPNameField(qtf);
			propagQtf = true;
		}
		else if (e instanceof NavigatorOperator) {
			Expression el = ((NavigatorOperator) e).getLeftExpression();
			propagQtf = el instanceof FieldOperand
				&& getPNameField(((FieldOperand) el).getField()).isClassPName();
			//The alias name is null because only the first element (the last
			// join) can be have an alias
			qtf = getField((NavigatorOperator) e,
				pnameFieldName,
				genClassElemOnly,
				genClassElemName,
				null);
			pnf = getPNameField(qtf);
		}
		else
			throw new MedorException(
				"Unknown left operand of the specified NavigatorOperand: " + e);

		//System.out.println("propagQtf=" + propagQtf);
		//System.out.println("qtf=" + qtf.getName());
		if (pnf == null)
			throw new MedorException("Impossible to find a PNameField");

		// 'qn' is the QueryTree of 'qtf"
		QueryTree qt = qtf.getQueryTree();

		if (!pnf.isClassPName()) {
			Reference ref = pnf.getReference();

			Class jormClass = null;
			if (ref instanceof ClassRef)
				jormClass = ((ClassRef) ref).getMOClass();
			else if (ref instanceof GenClassRef && ((GenClassRef)ref).isClassRef())
				jormClass = ((GenClassRef) ref).getClassRef().getMOClass();
			else
				throw new MedorException("Impossible to navigate with PName:" + ref);

			ClassExtent extent = null;
			String nodeName;
			if (aliasName != null) {
				nodeName = aliasName;
				//create the extent corresponding to the reference
				extent = new ClassExtent(
					jormClass, nodeName, pnameFieldName, false);
			} else {
				nodeName = jormClass.getFQName();
				//create the extent corresponding to the reference
				extent = new ClassExtent(
					jormClass,
					nodeName,
					new String[]{fieldName},
					true,
					pnameFieldName);

			}
			//System.out.println("ClassExtent of the class " + nodeName + " created");


			// pnf references the PNameField of the joined extent
			pnf = (PNameField) extent.getField(extent.getPNameFieldName());

			//System.out.println("Create Join");
			JoinProject join = new JoinProject("", null);

			//add the join condition: pnf should be equal to the
			//PName of the extent
			join.setQueryFilter(
				new Equal(
					new BasicFieldOperand(qtf),
					new BasicFieldOperand(pnf)
				)
			);
			//project in qt all fields of the QueryTree containing
			//the origin of the path expression
			Field[] fToAdd = qt.getTupleStructure().getFields();
			for (int i = 0; i < fToAdd.length; i++) {
				if (propagQtf || qtf != fToAdd[i]) {
					//System.out.println("\t-" + fToAdd[i].getName());
					join.addPropagatedField(
						fToAdd[i].getName(),
						fToAdd[i].getType(),
						new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
				}
				//else
				//System.out.println("\t-Do not propagate "  + fToAdd[i].getName());
			}

			// project in qt the fields corresponding to the end of the path
			// expression
			fToAdd = extent.getTupleStructure().getFields();
			for (int i = 0; i < fToAdd.length; i++) {
				//System.out.println("add " + fToAdd[i].getName());
				join.addPropagatedField(
					fToAdd[i].getName(),
					fToAdd[i].getType(),
					new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
			}
			qtf = (QueryTreeField)
				join.getField(extent.getFieldName(nodeName, fieldName));
			// qtf references a propagated field to the field with the name
			// 'fieldName'

			qt = join;
		}
		else {
			if (pnf.isInGenClass())
				throw new MedorException(
					"Impossible to navigate with a multi valued reference");

			//System.out.println("The field 'fieldName' must be found on 'qt'");
            // The field 'fieldName' must be found on 'qt'
			String fn = contains(qt.getTupleStructure(),
				pnf.getQueryTree().getName() + "." + fieldName);
            if (fn != null)
                qtf = (QueryTreeField) qt.getTupleStructure().getField(fn);
			else if (qt instanceof ClassExtent) {
				//In case where qt is a ClassExtent
				ClassExtent extent = (ClassExtent) qt;
				TypedElement te = extent.getMetaObject().getTypedElement(fieldName);
				if (te == null)
					throw new MedorException("The jorm class "
						+ extent.getMetaObject().getName()
						+ " does not provide a field " + fieldName);
				qtf = extent.addField(te);
			}
			else
				throw new MedorException(
					"No field " + fieldName + " found on the query tree " + qt);
		}
		pnf = getPNameField(qtf);
		//pnf can be null

		if (pnf != null && pnf.getReference() instanceof GenClassRef) {
			// Multivalued reference => create a GenClass node and the join

			GenClassRef gcr = (GenClassRef) pnf.getReference();
			String nodeName =
				((Class) gcr.getParent()).getFQName() + "." + gcr.getName();
			GenClassExtent extent =
				new GenClassExtent(gcr, nodeName, pnameFieldName);

			//'pnf' references the PNameField of the GenClass
			pnf = (PNameField) extent.getField(extent.getPNameFieldName());

			JoinProject join = new JoinProject("", null);

			//add the join condition: pnf should be equal to the
			//PName of the extent
			join.setQueryFilter(
				new Equal(
					new BasicFieldOperand(qtf),
					new BasicFieldOperand(pnf)
				)
			);

			//project in 'join' all fields of the inherited node
			Field[] fToAdd = qt.getTupleStructure().getFields();
			for (int i = 0; i < fToAdd.length; i++) {
				join.addPropagatedField(
					fToAdd[i].getName(),
					fToAdd[i].getType(),
					new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
			}

			if (genClassElemOnly) {
				Field elemToAdd = extent.getField(extent.getElementFieldName());
				qtf = join.addPropagatedField(genClassElemName,
					elemToAdd.getType(),
					new QueryTreeField[]{(QueryTreeField) elemToAdd});
			}
			else {
				//project in 'join' all fields of the GenClassExtent
				// and find the QueryTreeField which represents the element
				fToAdd = extent.getTupleStructure().getFields();
				for (int i = 0; i < fToAdd.length; i++) {
					QueryTreeField qtf2 = join.addPropagatedField(
						fToAdd[i].getName(),
						fToAdd[i].getType(),
						new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
					if (fToAdd[i].getName().equals(extent.getElementFieldName()))
						qtf = qtf2;
				}
				//qtf is the element field of the GenClass (PropagatedField to ...)
			}
		}
		//System.out.println("end: nav(" + e + "," + fieldName+")");
		return qtf;
	}

	private static PNameField getPNameField(Field field) throws MedorException {
		if (field instanceof PNameField)
			return (PNameField) field;
		else if (field instanceof PropagatedField) {
			return getPNameField(
				((PropagatedField) field).getPreviousFields()[0]);
		}
		else
			return null;
	}

	private static TypedElement getTypedElement(Class clazz, String fn) {
		int idx = fn.length();
		TypedElement te = null;
		te = clazz.getTypedElement(fn);
		while ((idx = fn.lastIndexOf(".", idx)) != -1 && te == null) {
			te = clazz.getTypedElement(fn.substring(idx + 1, fn.length()));
			idx--;
		}
		return te;
	}

	private static String contains(TupleStructure ts, String fn) {
		String res = (ts.contains(fn) ? fn : null);
		int idx = fn.lastIndexOf(".", fn.length());
        while (res == null && idx != -1) {
			res = fn.substring(idx + 1, fn.length());
			res = (ts.contains(res) ? res : null);
			idx = fn.lastIndexOf(".", idx - 1);
        }
        return res;
    }

	/**
	 * This method check if the required navigation already exists. In this case
	 * the initial QueryTree (at the start of the navigation) is returned.
	 * Otherwise the null is returned.
	 * @param nav is the NavigatorOperator which describes the navigation
	 * @param aliasName is the alias of the query tree which contains the
	 * searched field.
	 */
	private static QueryTree checkIfExist(NavigatorOperator nav, String aliasName)
		throws MedorException {
		if (aliasName == null)
			return null;
		String searchedFieldName = aliasName + "."
			+ ((Operand) nav.getRightExpression()).getString();
		QueryTree initialQT;
		QueryTreeField qtf;
		NavigatorOperator no = nav;
		Expression ex = no.getLeftExpression();
		while(ex instanceof NavigatorOperator) {
			no = (NavigatorOperator) ex;
			ex = no.getLeftExpression();
		}
		if (ex instanceof FieldOperand) {
			qtf = (QueryTreeField) ((FieldOperand) ex).getField();
			initialQT = qtf.getQueryTree();
			searchedFieldName = qtf.getName() + searchedFieldName;
		} else
			throw new MedorException(
				"Unknown left operand of the specified NavigatorOperand: " + ex);
		return (QueryTree)
			(initialQT.getTupleStructure().contains(searchedFieldName)
				? initialQT : null);
	}
}
