/**
 * copyright area
 */
package org.objectweb.medor.query.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author S.Chassande-Barrioz
 */
public class QueryTreePrinter {

	static int id = 0;

	public static void printQueryTree(QueryTree qt, Logger log, int level) {
		if (log != null && log.isLoggable(level)) {
			id = 0;
			printQT("", qt, log, level);
		}
	}

	public static void printQueryTree(QueryTree qt, Logger log) {
		if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
			id = 0;
			printQT("", qt, log, BasicLevel.DEBUG);
		}
	}

	private static void printQT(String prefix, QueryTree qt, Logger log, int level) {
		id++;
		int myId = id;
		log.log(level, prefix + "*Node number " + myId);
		log.log(level, prefix + " - Class " + qt);
		log.log(level, prefix + " - Name " + qt.getName());
		if (qt instanceof FilteredQueryTree) {
			Expression e = ((FilteredQueryTree) qt).getQueryFilter();
			if (e != null)
				log.log(level, prefix + " -Filter: " + ExpressionPrinter.e2str(e));
		}
		if (qt instanceof RdbQueryLeaf) {
			try {
				log.log(level, prefix + " - sqlRequest: " + ((RdbQueryLeaf) qt).getSqlRequest());
			} catch (MedorException e) {
			}
			if (qt instanceof RdbExpQueryLeaf) {
				QualifiedTable[] qts = ((RdbExpQueryLeaf) qt).getQualifiedTables();
				for (int i = 0; i < qts.length; i++) {
					log.log(level, prefix + " - tableName[" + i + "]:"
						+ qts[i].getTableName() + " as " + qts[i].getAliasName());
				}
			}
		}
		log.log(level, prefix + " -Field names:");
		Field[] fs = qt.getTupleStructure().getFields();
		String msg = null;
		for (int i = 0; i < fs.length; i++) {
			msg = prefix + "\t-" + fs[i].getName();
			if (fs[i] instanceof RdbField) {
				msg += "(" + ((RdbField) fs[i]).getColumnName();
				if (fs[i] instanceof RdbExpField)
					msg += " in " + ((RdbExpField) fs[i]).getTable().getTableName();
				msg += ")";
			} else if (fs[i] instanceof CalculatedField) {
				msg += " expression=" + ExpressionPrinter.e2str(((CalculatedField) fs[i]).getExpression());
			} else if (fs[i] instanceof PropagatedField) {
				Field[] previous = ((PropagatedField) fs[i]).getPreviousFields();
				msg += " previous=";
				for (int j = 0; j < previous.length; j++)
					msg += previous[j].getName() + " ";
			}
			log.log(level, msg);
		}
		log.log(level, "");
		if (qt instanceof QueryNode) {
			for (Iterator it = getChildren((QueryNode) qt).iterator(); it.hasNext();) {
				log.log(level, prefix + " -Children of " + myId);
				log.log(level, prefix + "\t---------------");
				QueryTree qtchild = (QueryTree) it.next();
				if (qtchild == null)
					log.log(level, prefix + "\tNull child !!!!!!!");
				else
					printQT(prefix + "\t", qtchild, log, level);
			}
		}
	}

	/**
	 * Updates the list of children by screening all Fields and the
	 * query filter.
	 */
	public static ArrayList getChildren(QueryNode qn) {
		ArrayList children = new ArrayList();
		Field[] fs = qn.getTupleStructure().getFields();
		//add children from Fields
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				Field[] anc = ((PropagatedField) fs[i]).getPreviousFields();
				for (int j = 0; j < anc.length; j++) {
					QueryTree qt = ((QueryTreeField) anc[j]).getQueryTree();
					//System.log.log(level, "PropagatedField " + fs[i].getName() + ": anc["+j+"]=" + anc[j].getName() + " , qt=" + qt);
					if (!children.contains(qt)) {
						children.add(qt);
					}
				}
			} else if (fs[i] instanceof NestedField) {
				Field[] grouped = ((NestedField) fs[i]).getFields();
				for (int j = 0; j < grouped.length; j++) {
					QueryTree qt = ((QueryTreeField) grouped[j]).getQueryTree();
					//System.log.log(level, "NestedField " + fs[i].getName() + ": grouped["+j+"]=" + grouped[j].getName() + " , qt=" + qt);
					if (!children.contains(qt)) {
						children.add(qt);
					}
				}
			} else if (fs[i] instanceof CalculatedField) {
				addChildrenFromExpression(children, ((CalculatedField) fs[i]).getExpression());
			}
		}
		addChildrenFromExpression(children, qn.getQueryFilter());
		return children;
	}

	/**
	 * Parses an Expression and adds the QueryTrees of the FieldOperands to
	 * the children of the current QueryNode.
	 * @param exp is the Expression to be parsed.
	 */
	private static void addChildrenFromExpression(ArrayList children, Expression exp) {
		if (exp instanceof FieldOperand) {
			Field f = ((FieldOperand) exp).getField();
			QueryTree qt = ((QueryTreeField) f).getQueryTree();
			//System.log.log(level, "fieldOperand " + f.getName() + " , qt=" + qt);
			if (!children.contains(qt)) {
				children.add(qt);
			}
		} else if (exp instanceof Operator) {
			for (int i = 0; i < ((Operator) exp).getOperandNumber(); i++) {
				addChildrenFromExpression(children, ((Operator) exp).getExpression(i));
			}
		}
	}

}
