/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.rdb.api;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.query.api.FilteredQueryTree;

/**
 * An RdbExpQueryLeaf is a particular type of RdbQueryLeaf for which the
 * associated SQL query if not given as a SQL String, but is computed from
 * an Expression (representing the filter) and from abstract information
 * made of QualifiedTables, representing the FROM part of the SQL query.<br>
 * The Medor optimizer can "reason" about RdbExpQueryLeaves, and merge several
 * such query leaves into one if they have the same DataStore.<br>
 * The life cycle of an RdbExpQueryLeaf is:
 * <ul><li>create the leaf (constructor)</li>
 * <li>add the fields (addRdbField)</li>
 * <li>add the expression which is built on the fields</li>
 * </ul>
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 *			Alexandre Lefebvre
 * 			</b></A>
 */
public interface RdbExpQueryLeaf extends RdbQueryLeaf, FilteredQueryTree {
    /**
     * Adds a Field to the RdbExpQueryLeaf
     * @param fieldName is the name of the Field to be added.
     * @param type is the PTyme of the Field to be added.
     * @param colName is the name of the column in the associated
     * QualifiedTable.
     * @param table is the QualifiedTable from which the Field is to be created.
     * @return the created RdbExpField.
     * @throws a QueryLeafException if the QualifiedTable is not in the
     * set of QualifiedTables associatged with the current RdbExpQueryLeaf.
     */
    public RdbExpField addRdbField(String fieldName,
                            PType type,
                            String colName,
                            QualifiedTable table) throws QueryLeafException;

	/**
     * Adds an RdbExpField to the RdbExpQueryLeaf. The QueryLeaf associated
	 * to the RdbExpField becomes the current RdbExpQueryLeaf.
     * @param rdbField is the RdbField to be added.
     */
    public RdbExpField addRdbField(RdbExpField rdbField)
		throws QueryLeafException;

    /**
     * Adds an RdbExpField from this object's TupleStructure to the list of the
     * OrderBy clause. We first must add this RdbExpField to the TupleStructure
     * of this Object before adding it to The list of the OrderBy clause.
     * @param orderBy of type of RdbExpField.
     * @throws a QuaryLeafException if this RdbExpField don't appears
     * in the list of the RdbExpField of the TupleStructure of this object
     * in the Select Clause.
     */
    public void addOrderByField(RdbExpField orderBy) throws QueryLeafException;

	/**
     * Adds an RdbExpField to the RdbExpQueryLeaf. The QueryLeaf associated
	 * to the RdbExpField becomes the current RdbExpQueryLeaf.
     * @param rdbField is the RdbField to be added.
     */
    public void removeRdbField(RdbExpField rdbField)
		throws QueryLeafException;

    /**
     * Gets the array of RdbExpField of the OrderBy Clause.
     * It returns null if there is no Order in the results
     * @return an array of RdbExpQueryLeaf
     */
    public RdbExpField[] getOrderByFields();

    /**
     * Sets the groupBy Field. This RdbExpField must be in The TupleStructure of
     * this QuaryLeaf Object.
     * @param groupBy : RdbExpField
     * @throws QueryLeafException if this RdbExpField don't appears in
     * the list of the fields which will be projected in Select clause
     */
    public void setGroupBy(RdbExpField groupBy) throws QueryLeafException;

     /**
     * Gets the RdbExpField of the GroupBy Clause.
     * It returns null if there is no Result Grouping
     * @return an RdbExpField
     */
    public RdbExpField getGroupByField();

    /**
     * Returns the associated QualifiedTables.
     * @returns the array of associated QualifiedTables.
     */
    public QualifiedTable[] getQualifiedTables();
}
