/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.rdb.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.filter.api.*;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.Like;
import org.objectweb.medor.filter.lib.InCollection;
import org.objectweb.medor.filter.lib.CollectionOperand;
import org.objectweb.medor.lib.BasicTupleStructure;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;

import java.util.Vector;
import java.util.Iterator;
import java.util.Collection;

/**
 * This class represents a QueryLeaf that maps onto a relational database store.
 * It implements the case where the relational query is provided as a set of
 * QualifiedTables and an Expression representing the filter.
 */
public class BasicRdbExpQueryLeaf extends BasicTupleStructure
        implements RdbExpQueryLeaf, TupleStructure {

    private final static String LAND = "&";
    private final static String AND = "AND";
    private final static String EQUAL = "=";
    private final static String LEQUAL = "==";
    private final static String LOR = "|";
    private final static String OR = "OR";

    Expression sqlFilter;
    DataStore ds;
    QualifiedTable[] from;
    String nodeName;
    Vector orderByFields = new Vector();
    RdbExpField groupByField;

    /**
     * Constructs a BasicRdbExpQueryLeaf from an Expression and a set of
     * QualifiedTables.
     * The SQL request (string) will be calculated.
     * @param ds is the DataStore associated to the QueryLeaf to be created
     * @param from is the array of QualifiedTables associated to the QueryLeaf
     * to be created
     * @throws a QueryLeafException if the QualifiedTables have all the same
     * names, and in this case also the same aliases.
     */
    public BasicRdbExpQueryLeaf(DataStore ds,
                                QualifiedTable[] from, String nodeName)
            throws QueryLeafException {
        super();
        this.nodeName = nodeName;
        this.ds = ds;
        //all QualifiedTables should have different names, and if not,
        //different aliases
        //TODO throw an exception if all QualifiedTables have the same name
        //and the same alias name
        this.from = from;
    }

    public String getName() {
        return nodeName;
    }

    /**
     * Adds a Field to the RdbExpQueryLeaf
     * @param fieldName is the name of the Field to be added.
     * @param type is the PTyme of the Field to be added.
     * @param colName is the name of the column in the associated
     * QualifiedTable.
     * @param table is the QualifiedTable from which the Field is to be created.
     * @return the created RdbExpField.
     */
    public RdbExpField addRdbField(String name,
                                   PType type,
                                   String colName,
                                   QualifiedTable table)
            throws QueryLeafException {
        //check that the parameter QualifiedTable is a valid QualifiedTable
        //in this RdbExpQueryLeaf
        QualifiedTable[] qts = this.getQualifiedTables();
        int i = 0;
        while (i < qts.length) {
            if (table == qts[i])
                break;
            else
                i++;
        }
        if (i == qts.length)
            throw new QueryLeafException("The parameter QualifiedTable is not a QualifiedTable of the current RdbExpQueryLeaf");
        else {
            RdbExpField f =
                    new BasicRdbExpField(name, type, colName, table, this);
            name2field.put(f.getName(), f);
            fields.add(f);
            return f;
        }
    }

	/**
     * Adds an RdbExpField to the RdbExpQueryLeaf
     * @param rdbField is the RdbField to be added.
     */
	public RdbExpField addRdbField(RdbExpField rdbField)
		throws QueryLeafException {
		rdbField.setQueryLeaf(this);
		name2field.put(rdbField.getName(), rdbField);
		fields.add(rdbField);
		return rdbField;
	}

	public void removeRdbField(RdbExpField rdbField)
		throws QueryLeafException {
		name2field.remove(rdbField.getName());
		fields.remove(rdbField);
	}

	/**
     * Builds and returns the SQL query as a String corresponding to the
     * current RdbExpQueryLeaf.
     * The SQL query is computed from the provided RdbExpFields (SELECT clause),
     * QualifiedTables (FROM clause) and Expression (WHERE clause).
     * @return the SQL query as a String.
     */
    public String getSqlRequest() throws MedorException {
        String query = getSelectList("SELECT ")
                + getFromList(" FROM ")
                + getWhereClause();
        if (orderByFields.size() > 0) {
            String orderByString = "";
            for (int cpt = 0; (cpt < orderByFields.size()); cpt++) {
                if (cpt > 1) orderByString = orderByString + ", ";
                orderByString = orderByString + getQualifiedFieldName((RdbExpField) (orderByFields.elementAt(cpt)));
            }
            query = query + " " + "ORDER BY " +
                    orderByString;
        }
        if (groupByField != null) {
            query = query + " " + "GROUP BY " +
                    getQualifiedFieldName(groupByField);
        }
        return query;
    }

    /**
     * It sets the filter associated to the RbdQueryLeaf. This filter is
     * represented by an Expression. Unlike the filter of a QueryNode, is
     * applies to its own Fields, instead of the Fields of its children.
     * @param exp is the Medor Expression representing the filter to be applied
     * to the relational source tables.
     */
    public void setQueryFilter(Expression exp) {
        this.sqlFilter = exp;
    }

    /**
     * It retrieves the filter associated to the RbdQueryLeaf.
     * @return the Medor Expression representing the filter to be applied to
     * the relational source tables.
     */
    public Expression getQueryFilter() {
        return this.sqlFilter;
    }

    /**
     * Returns the associated QualifiedTables.
     * @returns the array of associated QualifiedTables.
     */
    public QualifiedTable[] getQualifiedTables() {
        return this.from;
    }

    public DataStore getDataStore() {
        return ds;
    }

    public TupleStructure getTupleStructure() {
        return (TupleStructure) this;
    }

    public void addOrderByField(RdbExpField orderBy) throws QueryLeafException {
        if (orderByFields.contains(orderBy)) {
        } else if (fields.contains(orderBy)) {
            orderByFields.add(orderBy);
        } else
            throw new QueryLeafException("This RdbExpField don't exist in the TupleStructure of this RdbQueryLeaf");
    }

    public RdbExpField[] getOrderByFields() {
        return (RdbExpField[]) orderByFields.toArray(new RdbExpField[0]);
    }

    public void setGroupBy(RdbExpField groupBy) throws QueryLeafException {
        if (fields.contains(groupBy)) {
            groupByField = groupBy;
        } else
            throw new QueryLeafException("This RdbExpField don't exist in the TupleStructure of this RdbQueryLeaf");
    }

    public RdbExpField getGroupByField() {
        return groupByField;
    }

    /**
     * Builds the SELECT clause of the query (without "SELECT").
     * @return the list of qualified columns for the SELECT clause
     * @param selectList is the start of the SELECT clause to which the list
     * of qualified columns is appended.
     */
    private String getSelectList(String selectList) {
        for (int i = 0; i < fields.size(); i++) {
            selectList = selectList + getQualifiedFieldName(
                    ((RdbExpField) fields.get(i)));
            if (i < fields.size() - 1)
                selectList += ", ";
        }
        return selectList;
    }

    /**
     * Builds the qualified name of a RdbExpField.
     * @return the qualified name of an RdbExpField.
     * @param input is the input string to which the qualified name is appended
     * @param field is the RdbExpField for which the qualified name is built.
     */
    private String getQualifiedFieldName(RdbExpField field) {
        String res = null;
		if (field.getTable().getAliasName() == null)
            res = field.getTable().getTableName();
        else
            res = field.getTable().getAliasName();
        return res + "." + field.getColumnName();
    }

    /**
     * Builds the FROM clause of the query (without "FROM").
     * @return the list of qualified table names for the FROM clause
     * @param fromList is the start of the FROM clause to which the list of
     * tables is appended.
     */
    private String getFromList(String fromList) {
        for (int i = 0; i < from.length; i++) {
            fromList += from[i].getTableName();
            if (from[i].getAliasName() != null)
                fromList += " as " + from[i].getAliasName();
            if (i < from.length - 1)
                fromList += ", ";
        }
        return fromList;
    }

    /**
     * Builds the WHERE clause of the query (with the "WHERE") from the
     * Expression.
     * @return the WHERE clause of the query.
     */
    private String getWhereClause() throws MedorException {
        if (sqlFilter == null) {
            return "";
        } else
            return " WHERE " + parseExpression(sqlFilter);
    }

    /**
     * Computes a portion of SQL statement for the WHERE clause from a Medor
     * Expression.
     * @param input is the String to which the computed portion of the WHERE
     * clause should be appended.
     * @param exp is the Medor Expression from which to compute the portion of
     * the SQL WHERE clause.
     * @return the portion of the SQL WHERE clause corresponding to the
     * Expression.
     */
    private String parseExpression(Expression exp)
            throws MedorException {
        if (exp instanceof FieldOperand)
            return getQualifiedFieldName(
				((RdbExpField) ((FieldOperand) exp).getField()));

        else if (exp instanceof ParameterOperand)
            return "?";

        else if (exp instanceof BasicOperand)
            return "'" + ((BasicOperand) exp).getValueAsString() + "'";

		else if (exp instanceof Like) {
			Like l = (Like) exp;
			return
				"("
				+ parseExpression(l.getExpression(0))
				+ " LIKE "
				+ parseExpression(l.getExpression(1))
				+ (l.getOperandNumber() == 3
				? " ESCAPE " + parseExpression(l.getExpression(2)) + ")"
				: ")");
		} else if (exp instanceof InCollection) {
			InCollection bo  = (InCollection) exp;
			return
				"("
				+ parseExpression(bo.getLeftExpression())
				+ " "
				+ toSqlOperator(bo.getOperatorString())
				+ " "
				+ getSQLEnumeration(bo.getRightExpression(), ((InCollection)bo).getElemType())
				+ ")";
        } else if (exp instanceof BinaryOperator) {
			BinaryOperator bo  = (BinaryOperator) exp;
			return
				"("
				+ parseExpression(bo.getLeftExpression())
				+ " "
				+ toSqlOperator(bo.getOperatorString())
				+ " "
				+ parseExpression(bo.getRightExpression())
				+ ")";
        } else if (exp instanceof UnaryOperator) {
            ;
            return
				"("
				+ toSqlOperator(((Operator) exp).getOperatorString())
				+ " "
				+ parseExpression(((UnaryOperator) exp).getExpression())
				+ "(";
        } else
			throw new MedorException("Umanaged expression element: " + exp);
    }

	private String getSQLEnumeration(Expression e, PType elemType) throws MedorException {
		String result = "(";
		Iterator it = ((Collection)((CollectionOperand)e).getObject()).iterator();
		if (!it.hasNext())
			throw new MedorException("Empty collection for InCollection operator");
		while (it.hasNext()) {

			if (elemType.equals(PTypeSpace.STRING)) {
				result += "\"";
				result += it.next();
				result += "\"";
			}
			else
				result += it.next();
			if (it.hasNext())
				result += ", ";
		}
		return result += ")";
	}

    /**
     * Returns the SQL equivalent of the Java operator String.
     */
    private static String toSqlOperator(String in) {
        if (in.equals(BasicRdbExpQueryLeaf.LAND)) {
            return BasicRdbExpQueryLeaf.AND;
        } else if (in.equals(BasicRdbExpQueryLeaf.LOR)) {
            return BasicRdbExpQueryLeaf.OR;
        } else if (in.equals(BasicRdbExpQueryLeaf.LEQUAL)) {
            return BasicRdbExpQueryLeaf.EQUAL;
        } else
            return in;
    }
}
