/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.lib;


import java.util.LinkedList;
import java.util.List;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Collection;
import java.util.Iterator;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;

/**
 * This interface is an implementation of the TupleCollection interface using
 * Java collections. It can be used when we have obejcts fetched in the memory.
 */
public class ExplicitTupleCollection implements org.objectweb.medor.tuple.api.TupleCollection {

    public ExplicitTupleCollection(TupleStructure schema) throws MedorException {
        this.schema = schema;
        size = schema.getSize();
        tuples = new LinkedList();
        cursor = 0;
    }

    private TupleStructure schema;
    private List tuples;
    private int cursor;
    private int size;


    public synchronized TupleStructure getMetaData() {
        return schema;
    }

    public synchronized boolean isEmpty() {
        return (tuples.size() == 0);
    }


    public synchronized boolean isLast() throws MedorException {
        if ((cursor == -1) || (isEmpty()) || (cursor == tuples.size()))
            throw new MedorException(" No elements fetched in this tupleCollection");
        else
            return (card() == getRow());
    }


    public synchronized int card() {
        return tuples.size();
    }

    public synchronized boolean next() throws MedorException {
        if (!isEmpty()) {
            if (cursor < (card() - 1)) {
                cursor++;
                return true;
            } else {
                if (cursor == (card() - 1)) {
                    cursor++;
                    return true;
                } else
                    return false;
            }
        } else
            return false;
    }

    public synchronized boolean previous() throws MedorException {
        boolean arriere = true;
        if (!isEmpty()) {
            if (cursor != -1) {
                cursor--;
                if (cursor == -1)
                    arriere = false;
                else
                    arriere = true;
            }
        } else
            arriere = false;
        return arriere;
    }

    public synchronized boolean row(int row) throws MedorException {
        if ((row >= 1) && (row <= tuples.size())) {
            cursor = row - 1;
            return true;
        } else
            return false;
    }

    public synchronized void first() {
        if (!isEmpty()) cursor = 1;
    }

    public synchronized int getRow() {
        if ((isEmpty()) || (cursor == -1) || (cursor == card()))
            return cursor;
        else
            return cursor + 1;
    }

    /**
     * Insert the tuple T at the designated row number of the TupleCollection
     * @param i the first tuple is 1, the second is 2, ...
     * @param Tuple T
     * @throw MedorException if data source access error or invalid tuple number
     */
    public synchronized void insertTuple(int i, Tuple T) throws MedorException {
        if ((i >= 1) && (i <= card()))
            tuples.add((i - 1), T);
        else
            throw new MedorException("its impossible to insert in this cursor: " + i);
    }

    /**
     * Insert the tuple T at the last of this TupleCollection object
     * @param i the first tuple is 1, the second is 2, ...
     * @ a Tuple T
     * @throw MedorException if data source access error
     */
    public synchronized void insertTuple(Tuple T) throws MedorException {
        tuples.add(T);
    }

    /**
     * Remove the tuple of the designated row number from the TupleCollection object
     * @param i the first tuple is 1, the second is 2, ...
     * @throw MedorException if data source access error or invalid tuple number
     */
    public synchronized void removeTuple(int row) throws MedorException {
        if ((row >= 1) && (row <= card()))
            tuples.remove(row - 1);
        else
            throw new MedorException("its impossible to remove from this cursor");
    }

    /**
     * Returns the value of the designated row number of the TupleCollection
     * as a Tuple Object.
     * @param i the first tuple is 1, the second is 2, ...
     * @return a Tuple Object
     * @throw MedorException if data source access error
     */
    public synchronized Tuple getTuple(int i) throws MedorException {
        if ((i >= 1) && (i <= card())) {
            Object o = tuples.get(i - 1);
            Tuple upletI = (Tuple) o;
            return upletI;
        } else
            throw new MedorException("its impossible de get tuple from this cursor " + i);

    }

    public synchronized Tuple getTuple() throws MedorException {
        if ((!isEmpty()) && (cursor >= 0) && (cursor <= tuples.size())) {
            return (Tuple)tuples.get(cursor-1);

        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a Boolean of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a boolean value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized boolean getBoolean(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getBoolean(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a Byte of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a byte value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized byte getByte(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getByte(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.util.Date of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a Date value as SQL type
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized Date getDate(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple(cursor + 1);
            return upletI.getDate(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a double of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a double value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized double getDouble(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getDouble(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a float of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a float value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized float getFloat(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getFloat(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as an int of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return an integer value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized int getInt(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getInt(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a char of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a character value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized char getChar(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getChar(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a long of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a long value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized long getLong(int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple upletI = this.getTuple();
            return upletI.getLong(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.lang.Object of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return an Object value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized Object getObject(int i) throws MedorException {
        if ((!isEmpty()) && (cursor > -1) && (cursor < card())) {
            Tuple upletI = this.getTuple();
            return upletI.getObject(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a short of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a short value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized short getShort(int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple upletI = this.getTuple();
            return upletI.getShort(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.lang.String of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a String value
     * @throw MedorException if data source access error or invalid cursor
     */
    public synchronized String getString(int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple upletI = this.getTuple();
            return upletI.getString(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setBoolean(boolean x, int i) throws MedorException {
        //i represent the current row
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setBoolean(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }


    public synchronized void setByte(byte x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setByte(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setDate(Date x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setDate(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setDouble(double x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setDouble(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setFloat(float x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setFloat(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }


    public synchronized void setInt(int x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setInt(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /*public synchronized void setObject(Object x , int i) throws MedorException{
    	if ((!isEmpty()) && (cursor != -1) && (cursor != card())){
    		Tuple t = this.getTuple();
    		MemoryTuple upletI = (MemoryTuple)t;
	    	upletI.setJavaObject(x,i);
    		tuples.add(i,upletI);
    	}else throw new MedorException(TYPE_1,"No elements fetched in this tupleCollection");
    }*/

    public synchronized void setShort(short x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setShort(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setString(String x, int i) throws MedorException {
        if ((!isEmpty()) && (cursor != -1) && (cursor != card())) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setString(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.sql.Time of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a Time value as in SQL type
     * @throw MedorException if data source access error or invalid cursor
     */
    /* public synchronized Time getTime(int i) throws MedorException{
         if ((!isEmpty()) && (cursor != -1) && (cursor != card())){
             Tuple upletI = this.getTuple();
             return upletI.getTime(i);
         }else throw new MedorException("No elements fetched in this tupleCollection");
     }

      public synchronized Timestamp getTimestamp(int i) throws MedorException{
         if ((!isEmpty()) && (cursor != -1) && (cursor != card())){
         Tuple upletI = this.getTuple();
         return upletI.getTimestamp(i);
         }else throw new MedorException("No elements fetched in this tupleCollection");
      }

     public synchronized void setTime(Time x , int i) throws MedorException{
         if ((!isEmpty()) && (cursor != -1) && (cursor != card())){
             Tuple t = this.getTuple();
             MemoryTuple upletI = (MemoryTuple)t;
             upletI.setTime(x,i);
             tuples.add(i,upletI);
         }else throw new MedorException("No elements fetched in this tupleCollection");
     }

     public synchronized void setTimestamp(Timestamp x , int i) throws MedorException{
         if ((!isEmpty()) && (cursor != -1) && (cursor != card())){
             Tuple t = this.getTuple();
             MemoryTuple upletI = (MemoryTuple)t;
             upletI.setTimestamp(x,i);
             tuples.add(i,upletI);
         }else throw new MedorException("No elements fetched in this tupleCollection");
     }*/


    /**
     * Some operations of researchs and iterations.
     */

    public synchronized boolean contains(Tuple T) {
        return tuples.contains(T);
    }

    public synchronized void display() throws MedorException {
        TupleCollectionView frame = new TupleCollectionView(this);
        frame.pack();
        frame.setVisible(true);
    }

    public synchronized Iterator iteratorOf(int i) throws MedorException {
        Object o = getObject(i);
        if (o instanceof Collection) {
            Collection c = (Collection) o;
            return c.iterator();
        } else {
            throw new MedorException("the Attribute number " + i + " is not a collection Object");
        }
    }

    public synchronized Iterator iteratorAll() throws MedorException {
        return tuples.iterator();
    }

    /**
     * ------------------------------------------------------------------------
     * This methods will create a duplicated TC with an other cursor
     * ------------------------------------------------------------------------
     */

    private class DuplicatedETC implements TupleCollection {
        private int cursorD = 0;

        public synchronized TupleStructure getMetaData() {
            return schema;
        }


        public synchronized boolean isEmpty() {
            return (card() == 0);
        }


        public synchronized boolean isLast() throws MedorException {
            if ((cursorD == -1) || (isEmpty()) || (cursorD == card()))
                throw new MedorException(" No elements fetched in this tupleCollection");
            else
                return (card() == getRow());
        }


        public synchronized int card() {
            return tuples.size();
        }

        public synchronized boolean next() throws MedorException {
            if (!isEmpty()) {
                if (cursorD < (card() - 1)) {
                    cursorD++;
                    return true;
                } else {
                    if (cursorD == (card() - 1)) {
                        cursorD++;
                        return false;
                    } else
                        return false;
                }
            } else
                return false;
        }


        public synchronized boolean previous() throws MedorException {
            boolean arriere = true;
            if (!isEmpty()) {
                if (cursorD != -1) {
                    cursorD--;
                    if (cursorD == -1)
                        arriere = false;
                    else
                        arriere = true;
                }
            } else
                arriere = false;
            return arriere;
        }


        public synchronized boolean row(int row) throws MedorException {
            if ((row >= 1) && (row <= card())) {
                cursorD = row - 1;
                return true;
            } else
                return false;
        }


        public synchronized void first() {
            if (!isEmpty()) cursorD = 0;
        }


        public synchronized int getRow() {
            if ((isEmpty()) || (cursorD == -1) || (cursorD == card()))
                return 0;
            else
                return cursorD + 1;
        }


        public synchronized void insertTuple(int i, Tuple T) throws MedorException {
            if ((i >= 1) && (i <= card()))
                tuples.add((i - 1), T);
            else
                throw new MedorException("its impossible to insert in this cursorD: " + i);
        }


        public synchronized void insertTuple(Tuple T) throws MedorException {
            tuples.add(T);
        }


        public synchronized void removeTuple() throws MedorException {
            if ((cursor >= 0) && (cursor < card()))
                tuples.remove(cursor);
            else
                throw new MedorException("its impossible to remove from this cursorD");
        }


        public synchronized Tuple getTuple(int i) throws MedorException {
            if ((i >= 1) && (i <= card())) {
                Object o = tuples.get(i - 1);
                Tuple upletI = (Tuple) o;
                return upletI;
            } else
                throw new MedorException("its impossible de get tuple from this cursorD");
        }

        public synchronized Tuple getTuple() throws MedorException {
            if ((!isEmpty()) && (cursorD >= 0) && (cursorD < card())) {
                Object o = tuples.get(cursorD);
                Tuple upletI = (Tuple) o;
                return upletI;
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized boolean getBoolean(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getBoolean(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized byte getByte(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getByte(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized Date getDate(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple(cursorD + 1);
                return upletI.getDate(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized double getDouble(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getDouble(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized float getFloat(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getFloat(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized int getInt(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getInt(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized int getChar(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getChar(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized long getLong(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple upletI = this.getTuple();
                return upletI.getLong(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized Object getObject(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD > -1) && (cursorD < card())) {
                Tuple upletI = this.getTuple();
                return upletI.getObject(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized short getShort(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple upletI = this.getTuple();
                return upletI.getShort(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized String getString(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple upletI = this.getTuple();
                return upletI.getString(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        /**updating fields of tuplecollections
         */

        public synchronized void setBoolean(boolean x, int i) throws MedorException {
            //i represent the current row
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setBoolean(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized void setByte(byte x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setByte(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setDate(Date x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setDate(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setDouble(double x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setDouble(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setFloat(float x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setFloat(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }


        public synchronized void setInt(int x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setInt(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        /*public synchronized void setObject(Object x , int i) throws MedorException{
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())){
                Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple)t;
                upletI.setJavaObject(x,i);
                tuples.add(i,upletI);
            }else throw new MedorException("No elements fetched in this tupleCollection");
        }*/

        public synchronized void setShort(short x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                org.objectweb.medor.tuple.api.Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setShort(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setString(String x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                org.objectweb.medor.tuple.api.Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setString(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        /*public synchronized Time getTime(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple upletI = this.getTuple();
                return upletI.getTime(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized Timestamp getTimestamp(int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                Tuple upletI = this.getTuple();
                return upletI.getTimestamp(i);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setTime(Time x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                org.objectweb.medor.tuple.api.Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setTime(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }

        public synchronized void setTimestamp(Timestamp x, int i) throws MedorException {
            if ((!isEmpty()) && (cursorD != -1) && (cursorD != card())) {
                org.objectweb.medor.tuple.api.Tuple t = this.getTuple();
                MemoryTuple upletI = (MemoryTuple) t;
                upletI.setTimestamp(x, i);
                tuples.add(i, upletI);
            } else
                throw new MedorException("No elements fetched in this tupleCollection");
        }*/

        public synchronized boolean contains(org.objectweb.medor.tuple.api.Tuple T) {
            return tuples.contains(T);
        }

        public synchronized void display() throws MedorException {
            TupleCollectionView frame = new TupleCollectionView(this);
            frame.pack();
            frame.setVisible(true);
        }
    }

    /**
     * return a duplicate TupleCollection whith an other iterator and wich can manipulated independantly of
     * this one.
     * @return a TupleCollection
     */
    public synchronized TupleCollection getDuplicatedTupleCollection() {
        return new DuplicatedETC();
    }
}



