/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.query.jorm;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.NavigatorOperator;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.jorm.TestJormHelper;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Class;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class TestNavigatorNodeFactory extends TestJormPDHelper {

    public TestNavigatorNodeFactory(String n) {
        super(n);
    }


    public NavigatorOperator getNavigatorOperator(Field f, String[] path) throws MedorException {
        if (path.length == 0)
            return null;
        NavigatorOperator nav =
            new NavigatorOperator(new BasicFieldOperand(f), path[0]);
        for (int i = 1; i < path.length; i++) {
            nav = new NavigatorOperator(nav, path[i]);
        }
        return nav;
    }

    /**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a pname of the class
	 * - d.f1 where f1 is a primitive field.
	 * ==> The result is the same QueryTree
	 */
	public void test1a() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a pname of the class
	 * - d.e where e is a reference field.
	 * ==> The result is the same QueryTree.
	 */
    public void test1b() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a pname of the class
	 * - d.es where es is a GenClass of E.
	 * The result is a join with the GC node.
	 */
    public void test1c() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".es"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
            PNameField pnf = (PNameField)
                e.extent.getTupleStructure().getField(e.clazz.getFQName() + ".es");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent =
                new GenClassExtent(gcr, e.clazz.getFQName() + ".es", "object");

            JoinProject join = new JoinProject("", null);

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".es");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            //project in 'join' all fields of the GenClassExtent
            // and find the QueryTreeField which represents the element
            fToAdd = extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                QueryTreeField qtf2 = join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            TestQueryTreeHelper.equals("", join, result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public void test1d() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".es"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getIn(nav, null, "test1d");
            QueryTreePrinter.printQueryTree(result, logger);
            PNameField pnf = (PNameField)
                e.extent.getTupleStructure().getField(e.clazz.getFQName() + ".es");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent =
                new GenClassExtent(gcr, e.clazz.getFQName() + ".es", "object");

            JoinProject join = new JoinProject("", null);

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".es");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            //project in 'join' all fields of the GenClassExtent
            // and find the QueryTreeField which represents the element
			Field elemToAdd = extent.getField(extent.getElementFieldName());
			QueryTreeField qtf = join.addPropagatedField("test1d",
						elemToAdd.getType(),
						new QueryTreeField[]{(QueryTreeField) elemToAdd});

            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            TestQueryTreeHelper.equals("", join, result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a field of the class
	 * - d.f1 where f1 is a primitive field.
	 * ==> The result is a join with the D node
	 */
	public void test2a() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        try {
            QueryTreeField f = (QueryTreeField) e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[]{"f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
            QueryTreePrinter.printQueryTree(get2abExpected(e, "f1"), logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "f1"), result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a field of the class
	 * - d.e where e is a reference field.
	 * ==> The result is a join with the D node
	 */
    public void test2b() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[]{"e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "e"), result);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public QueryTree get2abExpected(Example e, String fn) throws MedorException {
        Example ed = getExtentOfD(new String[]{fn});
        ClassExtent extent = ed.extent;
        JoinProject join = new JoinProject("", null);

        //project in 'join' all fields of the inherited node
        Field[] fToAdd = e.extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(
                fToAdd[i].getName(),
                fToAdd[i].getType(),
                new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
        }

		fToAdd = extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(
                fToAdd[i].getName(),
                fToAdd[i].getType(),
                new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
        }

        join.setQueryFilter(
            new Equal(
                new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() +".d")),
                new BasicFieldOperand(extent.getField(ed.clazz.getFQName() + ".object"))
            )
        );
        return join;
    }

	/**
	 * Test the NavigatorNodeFactory.getNavigatorNode method.
	 * - start with a field of the class
	 * - d.es where es is a GenClass of E.
	 * The result is a join(join(qt, node(D), node(GC)).
	 */
	public void test2c() {
		TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
		try {
			Field f = e.extent.getField(e.clazz.getFQName() + ".d");
			String[] path = new String[]{"es"};
			NavigatorOperator nav = getNavigatorOperator(f, path);
			result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
			QueryTreePrinter.printQueryTree(result, logger);
		}
		catch (MedorException me) {
			me.printStackTrace();
            fail(me.getMessage());
		}
	}

	//A, b.c.d.f1
	public void test3a() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
	//A, b.c.d.e
	public void test3b() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
	//A, [A.b, c, d, es]
	public void test3c() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "es"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}

    //A, [c.d.es]
	public void test4a() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
    //A, [c.d.es]
	public void test4b() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
    //A, [c.d.es]
	public void test4c() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "es"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.getNavigatorNode(nav, null, null);
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
}
